"""Module to manage AiiDA codes."""

from subprocess import check_output

import ipywidgets as ipw
from IPython.display import clear_output
from traitlets import Bool, Dict, Instance, Unicode, Union, dlink, link, validate

from aiida.orm import Code, Computer, QueryBuilder, User
from aiida.plugins.entry_point import get_entry_point_names
from aiidalab_widgets_base.computers import ComputerDropdown


class CodeDropdown(ipw.VBox):
    """Code selection widget.
    Attributes:

    selected_code(Unicode or Code): Trait that points to the selected Code instance.
    It can be set either to an AiiDA Code instance or to a code label (will automatically
    be replaced by the corresponding Code instance). It is linked to the 'value' trait of
    the `self.dropdown` widget.

    codes(Dict): Trait that contains a dictionary (label => Code instance) for all
    codes found in the AiiDA database for the selected plugin. It is linked
    to the 'options' trait of the `self.dropdown` widget.

    allow_hidden_codes(Bool): Trait that defines whether to show hidden codes or not.

    allow_disabled_computers(Bool): Trait that defines whether to show codes on disabled
    computers.
    """
    selected_code = Union([Unicode(), Instance(Code)], allow_none=True)
    codes = Dict(allow_none=True)
    allow_hidden_codes = Bool(False)
    allow_disabled_computers = Bool(False)

    def __init__(self, input_plugin, description='Select code:', path_to_root='../', **kwargs):
        """Dropdown for Codes for one input plugin.

        input_plugin (str): Input plugin of codes to show.

        description (str): Description to display before the dropdown.
        """
        self.output = ipw.HTML()

        self.input_plugin = input_plugin

        self.dropdown = ipw.Dropdown(description=description, disabled=True, value=None)
        link((self, 'codes'), (self.dropdown, 'options'))
        link((self.dropdown, 'value'), (self, 'selected_code'))

        btn_refresh = ipw.Button(description="Refresh", layout=ipw.Layout(width="70px"))
        btn_refresh.on_click(self.refresh)

        self.observe(self.refresh, names=['allow_disabled_computers', 'allow_hidden_codes'])

        # Prepare URL parameters for the code setup.
        setup_code_params = {
            "label": input_plugin,
            "plugin": input_plugin,
        }

        for key, value in kwargs.pop('setup_code_params', {}).items():
            setup_code_params[key] = value

        ## For later: use base_url here, when it will be implemented in the appmode.
        url_string = f"<a href={path_to_root}aiidalab-widgets-base/setup_code.ipynb?"
        url_string += "&".join([f"{k}={v}".replace(' ', '%20') for k, v in setup_code_params.items()])
        url_string += ' target="_blank">Setup new code</a>'

        children = [ipw.HBox([self.dropdown, btn_refresh, ipw.HTML(value=url_string)]), self.output]

        super().__init__(children=children, **kwargs)

        self.refresh()

    def _get_codes(self):
        """Query the list of available codes."""

        user = User.objects.get_default()

        return {
            self._full_code_label(c[0]): c[0]
            for c in QueryBuilder().append(Code, filters={
                'attributes.input_plugin': self.input_plugin
            }).all()
            if c[0].computer.is_user_configured(user) and (self.allow_hidden_codes or not c[0].hidden) and
            (self.allow_disabled_computers or c[0].computer.is_user_enabled(user))
        }

    @staticmethod
    def _full_code_label(code):
        return f"{code.label}@{code.computer.name}"

    def refresh(self, _=None):
        """Refresh available codes.

        The job of this function is to look in AiiDA database, find available codes and
        put them in the dropdown attribute."""
        self.output.value = ''

        with self.hold_trait_notifications():
            self.dropdown.options = self._get_codes()
        if not self.dropdown.options:
            self.output.value = f"No codes found for input plugin '{self.input_plugin}'."
            self.dropdown.disabled = True
        else:
            self.dropdown.disabled = False
        self.dropdown.value = None

    @validate('selected_code')
    def _validate_selected_code(self, change):
        """If code is provided, set it as it is. If code's name is provided,
        select the code and set it."""
        code = change['value']
        self.output.value = ''

        # If code None, set value to None
        if code is None:
            return None

        # Check code by name.
        if isinstance(code, str):
            if code in self.codes:
                return self.codes[code]
            self.output.value = f"""No code named '<span style="color:red">{code}</span>'
            found in the AiiDA database."""

        # Check code by value.
        if isinstance(code, Code):
            label = self._full_code_label(code)
            if label in self.codes:
                return code
            self.output.value = f"""The code instance '<span style="color:red">{code}</span>'
            supplied was not found in the AiiDA database."""

        # This place will never be reached, because the trait's type is checked.
        return None


class AiiDACodeSetup(ipw.VBox):
    """Class that allows to setup AiiDA code"""
    label = Unicode()
    computer = Union([Unicode(), Instance(Computer)], allow_none=True)

    input_plugin = Unicode()
    description = Unicode()
    remote_abs_path = Unicode()
    prepend_text = Unicode()
    append_text = Unicode()

    def __init__(self, **kwargs):

        style = {"description_width": "200px"}

        # Code label.
        inp_label = ipw.Text(description="AiiDA code label:", layout=ipw.Layout(width="500px"), style=style)
        link((inp_label, 'value'), (self, 'label'))

        # Computer on which the code is installed. Two dlinks are needed to make sure we get a Computer instance.
        inp_computer = ComputerDropdown(layout={'margin': '0px 0px 0px 125px'})
        dlink((inp_computer, 'selected_computer'), (self, 'computer'))
        dlink((self, 'computer'), (inp_computer, 'selected_computer'))

        # Code plugin.
        inp_code_plugin = ipw.Dropdown(options=sorted(get_entry_point_names('aiida.calculations')),
                                       description="Code plugin:",
                                       layout=ipw.Layout(width="500px"),
                                       style=style)
        link((inp_code_plugin, 'value'), (self, 'input_plugin'))

        # Code description.
        inp_description = ipw.Text(placeholder='No description (yet)',
                                   description="Code description:",
                                   layout=ipw.Layout(width="500px"),
                                   style=style)
        link((inp_description, 'value'), (self, 'description'))

        inp_abs_path = ipw.Text(placeholder='/path/to/executable',
                                description="Absolute path to executable:",
                                layout=ipw.Layout(width="500px"),
                                style=style)
        link((inp_abs_path, 'value'), (self, 'remote_abs_path'))

        inp_prepend_text = ipw.Textarea(placeholder='Text to prepend to each command execution',
                                        description='Prepend text:',
                                        layout=ipw.Layout(width="400px"))
        link((inp_prepend_text, 'value'), (self, 'prepend_text'))

        inp_append_text = ipw.Textarea(placeholder='Text to append to each command execution',
                                       description='Append text:',
                                       layout=ipw.Layout(width="400px"))
        link((inp_append_text, 'value'), (self, 'append_text'))

        btn_setup_code = ipw.Button(description="Setup code")
        btn_setup_code.on_click(self._setup_code)
        self._setup_code_out = ipw.Output()
        children = [
            ipw.HBox([
                ipw.VBox([inp_label, inp_computer, inp_code_plugin, inp_description, inp_abs_path]),
                ipw.VBox([inp_prepend_text, inp_append_text])
            ]),
            btn_setup_code,
            self._setup_code_out,
        ]
        super(AiiDACodeSetup, self).__init__(children, **kwargs)

    def _setup_code(self, _=None):
        """Setup an AiiDA code."""
        with self._setup_code_out:
            clear_output()
            if self.label is None:
                print("You did not specify code label")
                return
            if not self.remote_abs_path:
                print("You did not specify absolute path to the executable")
                return
            if self.exists():
                print(f"Code {self.label}@{self.computer.name} already exists")
                return
            code = Code(remote_computer_exec=(self.computer, self.remote_abs_path))
            code.label = self.label
            code.description = self.description
            code.set_input_plugin_name(self.input_plugin)
            code.set_prepend_text(self.prepend_text)
            code.set_append_text(self.append_text)
            code.store()
            code.reveal()
            full_string = f"{self.label}@{self.computer.name}"
            print(check_output(['verdi', 'code', 'show', full_string]).decode('utf-8'))

    def exists(self):
        """Returns True if the code exists, returns False otherwise."""
        from aiida.common import NotExistent, MultipleObjectsError
        try:
            Code.get_from_string(f"{self.label}@{self.computer.name}")
            return True
        except MultipleObjectsError:
            return True
        except NotExistent:
            return False



"""Reusable widgets for AiiDAlab applications."""
# pylint: disable=unused-import,wrong-import-position
from aiida import load_profile
load_profile()

from .codes import CodeDropdown, AiiDACodeSetup
from .computers import SshComputerSetup
from .computers import AiidaComputerSetup
from .computers import ComputerDropdown
from .databases import CodQueryWidget, CodeDatabaseWidget, ComputerDatabaseWidget, OptimadeQueryWidget
from .export import ExportButtonWidget
from .process import ProgressBarWidget, ProcessFollowerWidget, ProcessInputsWidget, ProcessOutputsWidget
from .process import ProcessCallStackWidget, RunningCalcJobOutputWidget, SubmitButtonWidget, ProcessReportWidget
from .process import ProcessListWidget
from .structures import StructureManagerWidget
from .structures import StructureBrowserWidget, StructureExamplesWidget, StructureUploadWidget, SmilesWidget
from .structures import BasicStructureEditor
from .structures_multi import MultiStructureUploadWidget
from .viewers import viewer

__version__ = "1.0.0b14"



"""Widgets to work with processes."""
# pylint: disable=no-self-use

import os
from inspect import isclass
from time import sleep
import warnings
import pandas as pd
import ipywidgets as ipw
from IPython.display import HTML, Javascript, clear_output, display
from traitlets import Instance, Int, List, Unicode, Union, default, observe, validate

# AiiDA imports.
from aiida.engine import submit, Process, ProcessBuilder
from aiida.orm import CalcFunctionNode, CalcJobNode, Node, ProcessNode, WorkChainNode, WorkFunctionNode, load_node
from aiida.cmdline.utils.common import get_calcjob_report, get_workchain_report, get_process_function_report
from aiida.cmdline.utils.ascii_vis import format_call_graph
from aiida.cmdline.utils.query.calculation import CalculationQueryBuilder
from aiida.common.exceptions import MultipleObjectsError, NotExistent, NotExistentAttributeError

# Local imports.
from .viewers import viewer


def get_running_calcs(process):
    """Takes a process and yeilds running children calculations."""

    # If a process is a running calculation - returning it
    if issubclass(type(process), CalcJobNode) and not process.is_sealed:
        yield process

    # If the process is a running work chain - returning its children
    elif issubclass(type(process), WorkChainNode) and not process.is_sealed:
        for out_link in process.get_outgoing():
            if isinstance(out_link.node, ProcessNode) and not out_link.node.is_sealed:
                yield from get_running_calcs(out_link.node)


class SubmitButtonWidget(ipw.VBox):
    """Submit button class that creates submit button jupyter widget."""
    process = Instance(ProcessNode, allow_none=True)

    def __init__(  # pylint: disable=too-many-arguments
            self,
            process_class,
            inputs_generator=None,
            input_dictionary_function=None,
            description="Submit",
            disable_after_submit=True,
            append_output=False,
            **kwargs):
        """Submit Button widget.

        process_class (Process): Process class to submit.

        inputs_generator (func): Function that returns inputs dictionary or inputs builder.

        input_dictionary_function (DEPRECATED): Function that generates input parameters dictionary.

        description (str): Description written on the submission button.

        disable_after_submit (bool): Whether to disable the button after the process was submitted.

        append_output (bool): Whether to clear widget output for each subsequent submission.
        """

        self.path_to_root = kwargs.get('path_to_root', '../')
        if isclass(process_class) and issubclass(process_class, Process):
            self._process_class = process_class
        else:
            raise ValueError(f"process_class argument must be a sublcass of {Process}, got {process_class}")

        # Handling the deprecation.
        if inputs_generator is None and input_dictionary_function is None:
            raise ValueError("The `inputs_generator` argument must be provided.")
        if inputs_generator and input_dictionary_function:
            raise ValueError("You provided both: `inputs_generator` and `input_dictionary_function` "
                             "arguments. Please provide `inpust_generator` only.")
        if input_dictionary_function:
            inputs_generator = input_dictionary_function
            warnings.warn(("The `input_dictionary_function` argument is deprecated and "
                           "will be removed in the release 1.1 of the aiidalab-widgets-base package. "
                           "Please use the `inputs_generator` argument instead."), DeprecationWarning)

        # Checking if the inputs generator is
        if callable(inputs_generator):
            self.inputs_generator = inputs_generator
        else:
            raise ValueError("The `inputs_generator` argument must be a function that "
                             f"returns input dictionary, got {inputs_generator}")

        self.disable_after_submit = disable_after_submit
        self.append_output = append_output

        self.btn_submit = ipw.Button(description=description, disabled=False)
        self.btn_submit.on_click(self.on_btn_submit_press)
        self.submit_out = ipw.HTML('')
        self._run_after_submitted = []

        super().__init__(children=[self.btn_submit, self.submit_out])

    def on_click(self, function):
        self.btn_submit.on_click(function)

    def on_btn_submit_press(self, _=None):
        """When submit button is pressed."""

        if not self.append_output:
            self.submit_out.value = ''

        inputs = self.inputs_generator()
        if inputs is None:
            if self.append_output:
                self.submit_out.value += "SubmitButtonWidget: did not recieve the process inputs.<br>"
            else:
                self.submit_out.value = "SubmitButtonWidget: did not recieve the process inputs."
        else:
            if self.disable_after_submit:
                self.btn_submit.disabled = True
            if isinstance(inputs, ProcessBuilder):
                self.process = submit(inputs)
            else:
                self.process = submit(self._process_class, **inputs)

            if self.append_output:
                self.submit_out.value += f"""Submitted process {self.process}. Click
                <a href={self.path_to_root}aiidalab-widgets-base/process.ipynb?id={self.process.pk}
                target="_blank">here</a> to follow.<br>"""
            else:
                self.submit_out.value = f"""Submitted process {self.process}. Click
                <a href={self.path_to_root}aiidalab-widgets-base/process.ipynb?id={self.process.pk}
                target="_blank">here</a> to follow."""

            for func in self._run_after_submitted:
                func(self.process)

    def on_submitted(self, function):
        """Run functions after a process has been submitted successfully."""
        self._run_after_submitted.append(function)


class ProcessInputsWidget(ipw.VBox):
    """Widget to select and show process inputs."""

    process = Instance(ProcessNode, allow_none=True)

    def __init__(self, process=None, **kwargs):
        self.process = process
        self.output = ipw.Output()
        self.info = ipw.HTML()
        inputs_list = [(l.title(), l) for l in self.process.inputs] if self.process else []
        inputs = ipw.Dropdown(
            options=[('Select input', '')] + inputs_list,
            description='Select input:',
            style={'description_width': 'initial'},
            disabled=False,
        )
        inputs.observe(self.show_selected_input, names=['value'])
        super().__init__(children=[ipw.HBox([inputs, self.info]), self.output], **kwargs)

    def show_selected_input(self, change=None):
        """Function that displays process inputs selected in the `inputs` Dropdown widget."""
        with self.output:
            self.info.value = ''
            clear_output()
            if change['new']:
                selected_input = self.process.inputs[change['new']]
                self.info.value = "PK: {}".format(selected_input.id)
                display(viewer(selected_input))


class ProcessOutputsWidget(ipw.VBox):
    """Widget to select and show process outputs."""
    process = Instance(ProcessNode, allow_none=True)

    def __init__(self, process=None, **kwargs):
        self.process = process
        self.output = ipw.Output()
        self.info = ipw.HTML()
        outputs_list = [(l.title(), l) for l in self.process.outputs] if self.process else []
        outputs = ipw.Dropdown(
            options=[('Select output', '')] + outputs_list,
            label='Select output',
            description='Select outputs:',
            style={'description_width': 'initial'},
            disabled=False,
        )
        outputs.observe(self.show_selected_output, names=['value'])
        super().__init__(children=[ipw.HBox([outputs, self.info]), self.output], **kwargs)

    def show_selected_output(self, change=None):
        """Function that displays process output selected in the `outputs` Dropdown widget."""
        with self.output:
            self.info.value = ''
            clear_output()
            if change['new']:
                selected_output = self.process.outputs[change['new']]
                self.info.value = "PK: {}".format(selected_output.id)
                display(viewer(selected_output))


class ProcessFollowerWidget(ipw.VBox):
    """A Widget that follows a process until finished."""
    process = Instance(ProcessNode, allow_none=True)

    def __init__(self, process=None, followers=None, update_interval=0.1, path_to_root='../', **kwargs):
        """Initiate all the followers."""
        self.process = process
        self._run_after_completed = []
        self.update_interval = update_interval
        self.followers = []
        if followers is not None:
            for follower in followers:
                follower.process = self.process
                follower.path_to_root = path_to_root
                self.followers.append(ipw.VBox([
                    ipw.HTML("<h2><b>{}</b></h2>".format(follower.title)),
                    follower,
                ]))
        self.update()
        self.output = ipw.HTML()
        super().__init__(children=[self.output] + self.followers, **kwargs)

    def update(self):
        for follower in self.followers:
            follower.children[1].update()

    def _follow(self):
        """Periodically update all followers while the process is running."""
        while not self.process.is_sealed:
            self.update()
            sleep(self.update_interval)
        self.update()  # Update the state for the last time to be 100% sure.

        # Call functions to be run after the process is completed.
        for func in self._run_after_completed:
            func(self.process)

    def follow(self, detach=False):
        """Initiate following the process with or without blocking."""
        if self.process is None:
            self.output.value = """<font color="red"> ProcessFollowerWidget: process
            is set to 'None', nothing to follow. </font>"""
            return
        self.output.value = ''

        if detach:
            import threading
            update_state = threading.Thread(target=self._follow)
            update_state.start()
        else:
            self._follow()

    def on_completed(self, function):
        """Run functions after a process has been completed."""
        self._run_after_completed.append(function)


class ProcessReportWidget(ipw.HTML):
    """Widget that shows process report."""
    process = Instance(ProcessNode, allow_none=True)

    def __init__(self, title="Process Report", **kwargs):
        self.title = title
        self.max_depth = None
        self.indent_size = 2
        self.levelname = 'REPORT'
        self.update()
        super().__init__(**kwargs)

    def update(self):
        """Update report that is shown."""
        if self.process is None:
            return

        if isinstance(self.process, CalcJobNode):
            string = get_calcjob_report(self.process)
        elif isinstance(self.process, WorkChainNode):
            string = get_workchain_report(self.process, self.levelname, self.indent_size, self.max_depth)
        elif isinstance(self.process, (CalcFunctionNode, WorkFunctionNode)):
            string = get_process_function_report(self.process)
        else:
            string = 'Nothing to show for node type {}'.format(self.process.__class__)
        self.value = string.replace('\n', '<br/>')


class ProcessCallStackWidget(ipw.HTML):
    """Widget that shows process call stack."""
    process = Instance(ProcessNode, allow_none=True)

    def __init__(self, title="Process Call Stack", path_to_root='../', **kwargs):
        self.title = title
        self.path_to_root = path_to_root
        self.update()
        super().__init__(**kwargs)

    def update(self):
        """Update the call stack that is shown."""
        if self.process is None:
            return
        string = format_call_graph(self.process, info_fn=self.calc_info)
        self.value = string.replace('\n', '<br/>').replace(' ', '&nbsp;').replace('#space#', ' ')

    def calc_info(self, node):
        """Return a string with the summary of the state of a CalculationNode."""

        if not isinstance(node, ProcessNode):
            raise TypeError('Unknown type: {}'.format(type(node)))

        process_state = node.process_state.value.capitalize()
        pk = """<a#space#href={0}aiidalab-widgets-base/process.ipynb?id={1}#space#target="_blank">{1}</a>""".format(
            self.path_to_root, node.pk)

        if node.exit_status is not None:
            string = '{}<{}> {} [{}]'.format(node.process_label, pk, process_state, node.exit_status)
        else:
            string = '{}<{}> {}'.format(node.process_label, pk, process_state)

        if isinstance(node, WorkChainNode) and node.stepper_state_info:
            string += ' [{}]'.format(node.stepper_state_info)
        return string


class ProgressBarWidget(ipw.VBox):
    """A bar showing the proggress of a process."""
    process = Instance(ProcessNode, allow_none=True)

    def __init__(self, title="Progress Bar", **kwargs):
        """Initialize ProgressBarWidget."""

        self.title = title
        self.correspondance = {
            "created": 0,
            "running": 1,
            "waiting": 1,
            "killed": 2,
            "excepted": 2,
            "finished": 2,
        }
        self.progress_bar = ipw.IntProgress(
            value=0,
            min=0,
            max=2,
            step=1,
            description='Progress:',
            bar_style='warning',  # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal',
            layout={'width': '800px'})
        self.state = ipw.HTML(description="Calculation state:", value='Created', style={'description_width': 'initial'})
        super().__init__(children=[self.progress_bar, self.state], **kwargs)

    def update(self):
        """Update the bar."""
        if self.process is None:
            return
        self.progress_bar.value = self.correspondance[self.current_state]
        if self.current_state == 'finished':
            self.progress_bar.bar_style = 'success'
        elif self.current_state in ["killed", "excepted"]:
            self.progress_bar.bar_style = 'danger'
        else:
            self.progress_bar.bar_style = 'info'
        self.state.value = self.current_state.capitalize()

    @property
    def current_state(self):
        return self.process.process_state.value


class CalcJobOutputWidget(ipw.Textarea):
    """Output of a calculation."""
    calculation = Instance(CalcJobNode, allow_none=True)

    def __init__(self, **kwargs):
        default_params = {
            "value": "",
            "placeholder": "Calculation output will appear here",
            "description": "Calculation output:",
            "layout": {
                'width': "900px",
                'height': '300px'
            },
            "disabled": False,
            "style": {
                'description_width': 'initial'
            }
        }
        default_params.update(kwargs)
        self.output = []

        # Hack to make font monospace. As far as I am aware, currently there are no better ways.
        display(HTML("<style>textarea, input { font-family: monospace; }</style>"))

        super().__init__(**default_params)

    @observe('calculation')
    def _change_calculation(self, _=None):
        """Reset things if the observed calculation has changed."""
        self.output = []
        self.value = ''

    def update(self):
        """Update the displayed output and scroll to its end.

        NOTE: when this widgets is called by ProcessFollowerWidget in non-blocking manner
        the auto-scrolling won't work. There used to be a function for the Textarea widget,
        but it didn't work properly and got removed. For more information please visit:
        https://github.com/jupyter-widgets/ipywidgets/issues/1815"""

        if self.calculation is None:
            return

        try:
            output_file_path = os.path.join(self.calculation.outputs.remote_folder.get_remote_path(),
                                            self.calculation.attributes['output_filename'])
        except KeyError:
            self.placeholder = "The `output_filename` attribute is not set for " \
            f"{self.calculation.process_class}. Nothing to show."
        except NotExistentAttributeError:
            self.placeholder = "The object `remote_folder` was not found among the process outputs. " \
            "Nothing to show."
        else:
            if os.path.exists(output_file_path):
                with open(output_file_path) as fobj:
                    difference = fobj.readlines()[len(self.output):-1]  # Only adding the difference
                    self.output += difference
                    self.value += ''.join(difference)

        # Auto scroll down. Doesn't work in detached mode.
        # Also a hack as it is applied to all the textareas
        display(
            Javascript("""
            $('textarea').each(function(){
                // get the id
                var id_value = $(this).attr('id');

                if (typeof id_value !== 'undefined') {
                    // the variable is defined
                    var textarea = document.getElementById(id_value);
                    textarea.scrollTop = textarea.scrollHeight;
                }   
            });"""))


class RunningCalcJobOutputWidget(ipw.VBox):
    """Show an output of selected running child calculation."""
    process = Instance(ProcessNode, allow_none=True)

    def __init__(self, title="Running Job Output", **kwargs):
        self.title = title
        self.selection = ipw.Dropdown(description="Select calculation:",
                                      options={p.id: p for p in get_running_calcs(self.process)},
                                      style={'description_width': 'initial'})
        self.output = CalcJobOutputWidget()
        self.update()
        super().__init__(children=[self.selection, self.output], **kwargs)

    def update(self):
        """Update the displayed output."""
        if self.process is None:
            return
        with self.hold_trait_notifications():
            old_label = self.selection.label
            self.selection.options = {str(p.id): p for p in get_running_calcs(self.process)}
            # If the selection remains the same.
            if old_label in self.selection.options:
                self.label = old_label  # After changing options trait, the label and value traits might change as well.
            # If selection has changed
            else:
                self.output.calculation = self.selection.value
        self.output.update()


class ProcessListWidget(ipw.VBox):
    """List of AiiDA processes.

        past_days (int): Sumulations that were submitted in the last `past_days`.

        incoming_node (int, str, Node): Trait that takes node id or uuid and returns the node that must
        be among the input nodes of the process of interest.

        outgoing_node (int, str, Node): Trait that takes node id or uuid and returns the node that must
        be among the output nodes of the process of interest.

        process_states (list): List of allowed process states.

        process_label (str): Show process states of type `process_label`.

        description_contains (str): string that should be present in the description of a process node.

    """
    past_days = Int(7)
    incoming_node = Union([Int(), Unicode(), Instance(Node)], allow_none=True)
    outgoing_node = Union([Int(), Unicode(), Instance(Node)], allow_none=True)
    process_states = List()
    process_label = Unicode(allow_none=True)
    description_contains = Unicode(allow_none=True)

    def __init__(self, path_to_root='../', **kwargs):
        self.path_to_root = path_to_root
        self.table = ipw.HTML()
        pd.set_option('max_colwidth', 40)
        self.output = ipw.HTML()
        update_button = ipw.Button(description="Update now")
        update_button.on_click(self.update)
        self.update()
        super().__init__(children=[ipw.HBox([self.output, update_button]), self.table], **kwargs)

    def update(self, _=None):
        """Perform the query."""
        # Here we are defining properties of 'df' class (specified while exporting pandas table into html).
        # Since the exported object is nothing more than HTML table, all 'standard' HTML table settings
        # can be applied to it as well.
        # For more information on how to controle the table appearance please visit:
        # https://css-tricks.com/complete-guide-table-element/
        self.table.value = '''
        <style>
            .df { border: none; }
            .df tbody tr:nth-child(odd) { background-color: #e5e7e9; }
            .df tbody tr:nth-child(odd):hover { background-color:   #f5b7b1; }
            .df tbody tr:nth-child(even):hover { background-color:  #f5b7b1; }
            .df tbody td { min-width: 150px; text-align: center; border: none }
            .df th { text-align: center; border: none;  border-bottom: 1px solid black;}
        </style>
        '''
        builder = CalculationQueryBuilder()
        filters = builder.get_filters(all_entries=False,
                                      process_state=self.process_states,
                                      process_label=self.process_label,
                                      exit_status=None,
                                      failed=None)
        relationships = {}
        if self.incoming_node:
            relationships = {**relationships, **{'with_outgoing': self.incoming_node}}

        if self.outgoing_node:
            relationships = {**relationships, **{'with_incoming': self.outgoing_node}}

        query_set = builder.get_query_set(
            filters=filters,
            past_days=None if self.past_days < 0 else self.past_days,
            order_by={'ctime': 'desc'},
            relationships=relationships,
        )
        projected = builder.get_projected(
            query_set, projections=['pk', 'ctime', 'process_label', 'state', 'process_status', 'description'])
        dataf = pd.DataFrame(projected[1:], columns=projected[0])

        # Keep only process that contain the requested string in the description.
        if self.description_contains:
            dataf = dataf[dataf.Description.str.contains(self.description_contains)]

        self.output.value = "{} processes shown".format(len(dataf))

        # Add HTML links.
        dataf['PK'] = dataf['PK'].apply(
            lambda x: """<a href={0}aiidalab-widgets-base/process.ipynb?id={1} target="_blank">{1}</a>""".format(
                self.path_to_root, x))
        self.table.value += dataf.to_html(classes='df', escape=False, index=False)

    @validate('incoming_node')
    def _validate_incoming_node(self, provided):
        """Validate incoming node. The function load_node takes care of managing ids and uuids."""
        if provided['value']:
            try:
                return load_node(provided['value'])
            except (MultipleObjectsError, NotExistent):
                return None
        return None

    @validate('outgoing_node')
    def _validate_outgoing_node(self, provided):
        """Validate outgoing node. The function load_node takes care of managing ids and uuids."""
        if provided['value']:
            try:
                return load_node(provided['value'])
            except (MultipleObjectsError, NotExistent):
                return None
        return None

    @default('process_label')
    def _default_process_label(self):
        return None

    @validate('process_label')
    def _validate_process_label(self, provided):
        if provided['value']:
            return provided['value']
        return None

    def _follow(self, update_interval):
        while True:
            self.update()
            sleep(update_interval)

    def start_autoupdate(self, update_interval=10):
        import threading
        update_state = threading.Thread(target=self._follow, args=(update_interval,))
        update_state.start()



"""Useful functions that provide access to some data. """
from ase import Atom, Atoms
import numpy as np
import ipywidgets as ipw

# The first atom is anchoring, so the new bond will be connecting it
# The direction of the new bond is (-1, -1, -1).
LIGANDS = {
    'Select ligand':
        0,
    'Methyl -CH3': [('C', 0, 0, 0), ('H', 0.23962342, -0.47699124, 0.78585262),
                    ('H', 0.78584986, 0.23962732, -0.47698795), ('H', -0.47699412, 0.78585121, 0.23962671)],
    'Methylene =CH2': [('C', 0, 0, 0), ('H', -0.39755349, 0.59174911, 0.62728004),
                       ('H', 0.94520686, -0.04409933, -0.07963039)],
    'Hydroxy -OH': [('O', 0, 0, 0), ('H', 0.87535922, -0.3881659, 0.06790889)],
    'Amine -NH2': [('N', 0, 0, 0), ('H', 0.7250916, -0.56270993, 0.42151063),
                   ('H', -0.56261958, 0.4215284, 0.72515241)],
}


class LigandSelectorWidget(ipw.Dropdown):
    """Class to select ligands that are returned as `Atoms` object"""

    def __init__(self, value=0, description="Select ligand", **kwargs):
        self.style = {'description_width': 'initial'}
        self.layout = {'width': 'initial'}
        super().__init__(value=value, description=description, options=LIGANDS, **kwargs)

    def rotate(self, align_to=(0, 0, 1), remove_anchor=False):
        """Rotate group in such a way that vector which was (-1,-1,-1)
        is alligned with align_to."""

        vect = np.array(align_to)
        norm = np.linalg.norm(vect)

        if self.value == 0:
            return None

        mol = Atoms()
        for atm in self.value:
            mol.append(Atom(atm[0], atm[1:]))

        # Bad cases.
        if norm == 0.0:
            vect = np.array((1, 1, 1)) / np.sqrt(3)
        else:
            vect = vect / norm

        mol.rotate((1, 1, 1), vect)

        if remove_anchor:
            del mol[0]

        return mol

    @property
    def anchoring_atom(self):
        """Return anchoring atom chemical symbol."""
        if self.value == 0:
            return None
        return self.value[0][0]



"""Jupyter viewers for AiiDA data objects."""
# pylint: disable=no-self-use

import base64
import warnings
import numpy as np
import ipywidgets as ipw
from IPython.display import display
import nglview
from ase import Atoms

from traitlets import Instance, Int, List, Union, default, link, observe, validate
from aiida.orm import Node

from .utils import string_range_to_list, list_to_string_range
from .misc import CopyToClipboardButton


def viewer(obj, downloadable=True, **kwargs):
    """Display AiiDA data types in Jupyter notebooks.

    :param downloadable: If True, add link/button to download the content of displayed AiiDA object.
    :type downloadable: bool

    Returns the object itself if the viewer wasn't found."""
    if not isinstance(obj, Node):  # only working with AiiDA nodes
        warnings.warn("This viewer works only with AiiDA objects, got {}".format(type(obj)))
        return obj

    try:
        _viewer = AIIDA_VIEWER_MAPPING[obj.node_type]
        return _viewer(obj, downloadable=downloadable, **kwargs)
    except (KeyError) as exc:
        if obj.node_type in str(exc):
            warnings.warn("Did not find an appropriate viewer for the {} object. Returning the object "
                          "itself.".format(type(obj)))
            return obj
        raise exc


class DictViewer(ipw.HTML):
    """Viewer class for Dict object.

    :param parameter: Dict object to be viewed
    :type parameter: Dict
    :param downloadable: If True, add link/button to download the content of the object
    :type downloadable: bool"""

    def __init__(self, parameter, downloadable=True, **kwargs):
        super().__init__(**kwargs)
        import pandas as pd
        # Here we are defining properties of 'df' class (specified while exporting pandas table into html).
        # Since the exported object is nothing more than HTML table, all 'standard' HTML table settings
        # can be applied to it as well.
        # For more information on how to controle the table appearance please visit:
        # https://css-tricks.com/complete-guide-table-element/
        self.value = '''
        <style>
            .df { border: none; }
            .df tbody tr:nth-child(odd) { background-color: #e5e7e9; }
            .df tbody tr:nth-child(odd):hover { background-color:   #f5b7b1; }
            .df tbody tr:nth-child(even):hover { background-color:  #f5b7b1; }
            .df tbody td { min-width: 300px; text-align: center; border: none }
            .df th { text-align: center; border: none;  border-bottom: 1px solid black;}
        </style>
        '''
        pd.set_option('max_colwidth', 40)
        dataf = pd.DataFrame([(key, value) for key, value in sorted(parameter.get_dict().items())],
                             columns=['Key', 'Value'])
        self.value += dataf.to_html(classes='df', index=False)  # specify that exported table belongs to 'df' class
        # this is used to setup table's appearance using CSS
        if downloadable:
            payload = base64.b64encode(dataf.to_csv(index=False).encode()).decode()
            fname = '{}.csv'.format(parameter.pk)
            to_add = """Download table in csv format: <a download="{filename}"
            href="data:text/csv;base64,{payload}" target="_blank">{title}</a>"""
            self.value += to_add.format(filename=fname, payload=payload, title=fname)


class _StructureDataBaseViewer(ipw.VBox):
    """Base viewer class for AiiDA structure or trajectory objects.

    :param configure_view: If True, add configuration tabs
    :type configure_view: bool"""

    selection = List(Int)
    supercell = List(Int)
    DEFAULT_SELECTION_OPACITY = 0.2
    DEFAULT_SELECTION_RADIUS = 6
    DEFAULT_SELECTION_COLOR = 'green'

    def __init__(self, configure_view=True, **kwargs):
        # Defining viewer box.

        # 1. Nglviwer
        self._viewer = nglview.NGLWidget()
        self._viewer.camera = 'orthographic'
        self._viewer.observe(self._on_atom_click, names='picked')
        self._viewer.stage.set_parameters(mouse_preset='pymol')

        # 2. Camera type.
        camera_type = ipw.ToggleButtons(options={
            'Orthographic': 'orthographic',
            'Perspective': 'perspective'
        },
                                        description='Camera type:',
                                        value='orthographic',
                                        layout={"align_self": "flex-end"},
                                        style={'button_width': '115.5px'},
                                        orientation='vertical')

        def change_camera(change):
            self._viewer.camera = change['new']

        camera_type.observe(change_camera, names="value")
        view_box = ipw.VBox([self._viewer, camera_type])

        # Constructing configuration box
        if configure_view:
            configuration_box = ipw.Tab(layout=ipw.Layout(flex='1 1 auto', width='auto'))
            configuration_box.children = [self._selection_tab(), self._appearance_tab(), self._download_tab()]
            for i, title in enumerate(["Selection", "Appearance", "Download"]):
                configuration_box.set_title(i, title)
            children = [ipw.HBox([view_box, configuration_box])]
            view_box.layout = {'width': "60%"}
        else:
            children = [view_box]

        if 'children' in kwargs:
            children += kwargs.pop('children')

        super().__init__(children, **kwargs)

    def _selection_tab(self):
        """Defining the selection tab."""

        # 1. Selected atoms.
        self._selected_atoms = ipw.Text(description='Selected atoms:', value='', style={'description_width': 'initial'})

        # 2. Copy to clipboard
        copy_to_clipboard = CopyToClipboardButton(description="Copy to clipboard")
        link((self._selected_atoms, 'value'), (copy_to_clipboard, 'value'))

        # 3. Informing about wrong syntax.
        self.wrong_syntax = ipw.HTML(
            value="""<i class="fa fa-times" style="color:red;font-size:2em;" ></i> wrong syntax""",
            layout={'visibility': 'hidden'})

        # 4. Button to clear selection.
        clear_selection = ipw.Button(description="Clear selection")
        clear_selection.on_click(lambda _: self.set_trait('selection', list()))  # lambda cannot contain assignments

        # 5. Button to apply selection
        apply_selection = ipw.Button(description="Apply selection")
        apply_selection.on_click(self.apply_selection)

        self.selection_info = ipw.HTML()

        return ipw.VBox([
            ipw.HBox([self._selected_atoms, self.wrong_syntax]),
            ipw.HBox([copy_to_clipboard, clear_selection, apply_selection]),
            self.selection_info,
        ])

    def _appearance_tab(self):
        """Defining the appearance tab."""

        # 1. Supercell
        def change_supercell(_=None):
            self.supercell = [_supercell[0].value, _supercell[1].value, _supercell[2].value]

        _supercell = [
            ipw.BoundedIntText(value=1, min=1, layout={"width": "30px"}),
            ipw.BoundedIntText(value=1, min=1, layout={"width": "30px"}),
            ipw.BoundedIntText(value=1, min=1, layout={"width": "30px"}),
        ]
        for elem in _supercell:
            elem.observe(change_supercell, names='value')
        supercell_selector = ipw.HBox([ipw.HTML(description="Super cell:")] + _supercell)

        # 2. Choose background color.
        background_color = ipw.ColorPicker(description="Background")
        link((background_color, 'value'), (self._viewer, 'background'))
        background_color.value = 'white'

        # 3. Center button.
        center_button = ipw.Button(description="Center")
        center_button.on_click(lambda c: self._viewer.center())

        return ipw.VBox([supercell_selector, background_color, center_button])

    def _download_tab(self):
        """Defining the download tab."""

        # 1. Choose download file format.
        self.file_format = ipw.Dropdown(options=['xyz', 'cif'], layout={"width": "200px"}, description="File format:")

        # 2. Download button.
        self.download_btn = ipw.Button(description="Download")
        self.download_btn.on_click(self.download)
        self.download_box = ipw.VBox(
            children=[ipw.Label("Download as file:"),
                      ipw.HBox([self.file_format, self.download_btn])])

        # 3. Screenshot button
        self.screenshot_btn = ipw.Button(description="Screenshot", icon='camera')
        self.screenshot_btn.on_click(lambda _: self._viewer.download_image())
        self.screenshot_box = ipw.VBox(children=[ipw.Label("Create a screenshot:"), self.screenshot_btn])

        return ipw.VBox([self.download_box, self.screenshot_box])

    def _on_atom_click(self, _=None):
        """Update selection when clicked on atom."""
        if 'atom1' not in self._viewer.picked.keys():
            return  # did not click on atom
        index = self._viewer.picked['atom1']['index']
        selection = self.selection.copy()

        if selection:
            if index not in selection:
                selection.append(index)
            else:
                selection.remove(index)
        else:
            selection = [index]

        self.selection = selection

    def highlight_atoms(self,
                        vis_list,
                        color=DEFAULT_SELECTION_COLOR,
                        size=DEFAULT_SELECTION_RADIUS,
                        opacity=DEFAULT_SELECTION_OPACITY):
        """Highlighting atoms according to the provided list."""
        if not hasattr(self._viewer, "component_0"):
            return
        self._viewer._remove_representations_by_name(repr_name='selected_atoms')  # pylint:disable=protected-access
        self._viewer.add_ball_and_stick(  # pylint:disable=no-member
            name="selected_atoms",
            selection=list() if vis_list is None else vis_list,
            color=color,
            aspectRatio=size,
            opacity=opacity)

    @default('supercell')
    def _default_supercell(self):
        return [1, 1, 1]

    @default('selection')
    def _default_selection(self):
        return list()

    @validate('selection')
    def _validate_selection(self, provided):
        return list(provided['value'])

    @observe('selection')
    def _observe_selection(self, _=None):
        self.highlight_atoms(self.selection)
        self._selected_atoms.value = list_to_string_range(self.selection, shift=1)

    def apply_selection(self, _=None):
        """Apply selection specified in the text field."""
        selection_string = self._selected_atoms.value
        expanded_selection, syntax_ok = string_range_to_list(self._selected_atoms.value, shift=-1)
        self.wrong_syntax.layout.visibility = 'hidden' if syntax_ok else 'visible'
        self.selection = expanded_selection
        self._selected_atoms.value = selection_string  # Keep the old string for further editing.

    def download(self, change=None):  # pylint: disable=unused-argument
        """Prepare a structure for downloading."""
        self._download(payload=self._prepare_payload(), filename='structure.' + self.file_format.value)

    @staticmethod
    def _download(payload, filename):
        """Download payload as a file named as filename."""
        from IPython.display import Javascript
        javas = Javascript("""
            var link = document.createElement('a');
            link.href = "data:;base64,{payload}"
            link.download = "{filename}"
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            """.format(payload=payload, filename=filename))
        display(javas)

    def _prepare_payload(self, file_format=None):
        """Prepare binary information."""
        from tempfile import NamedTemporaryFile
        file_format = file_format if file_format else self.file_format.value
        tmp = NamedTemporaryFile()
        self.structure.write(tmp.name, format=file_format)  # pylint: disable=no-member
        with open(tmp.name, 'rb') as raw:
            return base64.b64encode(raw.read()).decode()

    @property
    def thumbnail(self):
        return self._prepare_payload(file_format='png')


class StructureDataViewer(_StructureDataBaseViewer):
    """Viewer class for AiiDA structure objects.

    Attributes:
        structure (Atoms, StructureData, CifData): Trait that contains a structure object,
        which was initially provided to the viewer. It can be either directly set to an
        ASE Atoms object or to AiiDA structure object containing `get_ase()` method.

        displayed_structure (Atoms): Trait that contains a structure object that is
        currently displayed (super cell, for example). The trait is generated automatically
        and can't be set outside of the class.
    """
    structure = Union([Instance(Atoms), Instance(Node)], allow_none=True)
    displayed_structure = Instance(Atoms, allow_none=True, read_only=True)

    def __init__(self, structure=None, **kwargs):
        super().__init__(**kwargs)
        self.structure = structure
        #self.supercell.observe(self.repeat, names='value')

    @observe('supercell')
    def repeat(self, _=None):
        if self.structure is not None:
            self.set_trait('displayed_structure', self.structure.repeat(self.supercell))

    @validate('structure')
    def _valid_structure(self, change):  # pylint: disable=no-self-use
        """Update structure."""
        structure = change['value']

        if structure is None:
            return None  # if no structure provided, the rest of the code can be skipped

        if isinstance(structure, Atoms):
            return structure
        if isinstance(structure, Node):
            return structure.get_ase()
        raise ValueError("Unsupported type {}, structure must be one of the following types: "
                         "ASE Atoms object, AiiDA CifData or StructureData.")

    @observe('structure')
    def _update_displayed_structure(self, change):
        """Update displayed_structure trait after the structure trait has been modified."""
        # Remove the current structure(s) from the viewer.
        if change['new'] is not None:
            self.set_trait('displayed_structure', change['new'].repeat(self.supercell))
        else:
            self.set_trait('displayed_structure', None)

    @observe('displayed_structure')
    def _update_structure_viewer(self, change):
        """Update the view if displayed_structure trait was modified."""
        with self.hold_trait_notifications():
            for comp_id in self._viewer._ngl_component_ids:  # pylint: disable=protected-access
                self._viewer.remove_component(comp_id)
            self.selection = list()
            if change['new'] is not None:
                self._viewer.add_component(nglview.ASEStructure(change['new']))
                self._viewer.clear()
                self._viewer.add_ball_and_stick(aspectRatio=4)  # pylint: disable=no-member
                self._viewer.add_unitcell()  # pylint: disable=no-member

    def create_selection_info(self):
        """Create information to be displayed with selected atoms"""

        if not self.selection:
            return ''

        def print_pos(pos):
            return " ".join([str(i) for i in pos.round(2)])

        def add_info(indx, atom):
            return "Id: {}; Symbol: {}; Coordinates: ({})<br>".format(indx + 1, atom.symbol, print_pos(atom.position))

        # Find geometric center.
        geom_center = print_pos(np.average(self.structure[self.selection].get_positions(), axis=0))

        # Report coordinates.
        if len(self.selection) == 1:
            return add_info(self.selection[0], self.structure[self.selection[0]])

        # Report coordinates, distance and center.
        if len(self.selection) == 2:
            info = ""
            info += add_info(self.selection[0], self.structure[self.selection[0]])
            info += add_info(self.selection[1], self.structure[self.selection[1]])
            info += "Distance: {}<br>Geometric center: ({})".format(
                self.structure.get_distance(*self.selection).round(2), geom_center)
            return info

        # Report angle geometric center and normal.
        if len(self.selection) == 3:
            angle = self.structure.get_angle(*self.selection).round(2)
            normal = np.cross(*self.structure.get_distances(
                self.selection[1], [self.selection[0], self.selection[2]], mic=False, vector=True))
            normal = normal / np.linalg.norm(normal)
            return "{} atoms selected<br>Geometric center: ({})<br>Angle: {}<br>Normal: ({})".format(
                len(self.selection), geom_center, angle, print_pos(normal))

        # Report  dihedral angle and geometric center.
        if len(self.selection) == 4:
            dihedral = self.structure.get_dihedral(self.selection) * 180 / np.pi
            return "{} atoms selected<br>Geometric center: ({})<br>Dihedral angle: {}".format(
                len(self.selection), geom_center, dihedral.round(2))

        return "{} atoms selected<br>Geometric center: ({})".format(len(self.selection), geom_center)

    @observe('selection')
    def _observe_selection_2(self, _=None):
        self.selection_info.value = self.create_selection_info()


class FolderDataViewer(ipw.VBox):
    """Viewer class for FolderData object.

    :param folder: FolderData object to be viewed
    :type folder: FolderData
    :param downloadable: If True, add link/button to download the content of the selected file in the folder
    :type downloadable: bool"""

    def __init__(self, folder, downloadable=True, **kwargs):
        self._folder = folder
        self.files = ipw.Dropdown(
            options=[obj.name for obj in self._folder.list_objects()],
            description="Select file:",
        )
        self.text = ipw.Textarea(value="",
                                 description='File content:',
                                 layout={
                                     'width': "900px",
                                     'height': '300px'
                                 },
                                 disabled=False)
        self.change_file_view()
        self.files.observe(self.change_file_view, names='value')
        children = [self.files, self.text]
        if downloadable:
            self.download_btn = ipw.Button(description="Download")
            self.download_btn.on_click(self.download)
            children.append(self.download_btn)
        super().__init__(children, **kwargs)

    def change_file_view(self, change=None):  # pylint: disable=unused-argument
        with self._folder.open(self.files.value) as fobj:
            self.text.value = fobj.read()

    def download(self, change=None):  # pylint: disable=unused-argument
        """Prepare for downloading."""
        from IPython.display import Javascript

        payload = base64.b64encode(self._folder.get_object_content(self.files.value).encode()).decode()
        javas = Javascript("""
            var link = document.createElement('a');
            link.href = "data:;base64,{payload}"
            link.download = "{filename}"
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            """.format(payload=payload, filename=self.files.value))
        display(javas)


class BandsDataViewer(ipw.VBox):
    """Viewer class for BandsData object.

    :param bands: BandsData object to be viewed
    :type bands: BandsData"""

    def __init__(self, bands, **kwargs):
        from bokeh.io import show, output_notebook
        from bokeh.models import Span
        from bokeh.plotting import figure
        output_notebook(hide_banner=True)
        out = ipw.Output()
        with out:
            plot_info = bands._get_bandplot_data(cartesian=True, join_symbol="|")  # pylint: disable=protected-access
            # Extract relevant data
            y_data = plot_info['y'].transpose().tolist()
            x_data = [plot_info['x'] for i in range(len(y_data))]
            labels = plot_info['labels']
            # Create the figure
            plot = figure(y_axis_label='Dispersion ({})'.format(bands.units))
            plot.multi_line(x_data, y_data, line_width=2, line_color='red')  # pylint: disable=too-many-function-args
            plot.xaxis.ticker = [l[0] for l in labels]
            # This trick was suggested here: https://github.com/bokeh/bokeh/issues/8166#issuecomment-426124290
            plot.xaxis.major_label_overrides = {int(l[0]) if l[0].is_integer() else l[0]: l[1] for l in labels}
            # Add vertical lines
            plot.renderers.extend(
                [Span(location=l[0], dimension='height', line_color='black', line_width=3) for l in labels])
            show(plot)
        children = [out]
        super().__init__(children, **kwargs)


AIIDA_VIEWER_MAPPING = {
    'data.dict.Dict.': DictViewer,
    'data.structure.StructureData.': StructureDataViewer,
    'data.cif.CifData.': StructureDataViewer,
    'data.folder.FolderData.': FolderDataViewer,
    'data.array.bands.BandsData.': BandsDataViewer,
}



"""Module to provide functionality to import structures."""
# pylint: disable=no-self-use

import io
import datetime
from collections import OrderedDict
import numpy as np
import ipywidgets as ipw
from traitlets import Instance, Int, List, Unicode, Union, dlink, link, default, observe

# ASE imports
import ase
from ase import Atom, Atoms
from ase.data import chemical_symbols, covalent_radii

# AiiDA imports
from aiida.engine import calcfunction
from aiida.orm import CalcFunctionNode, CalcJobNode, Data, QueryBuilder, Node, WorkChainNode
from aiida.plugins import DataFactory

from sklearn.decomposition import PCA

# Local imports
from .utils import get_ase_from_file
from .viewers import StructureDataViewer
from .data import LigandSelectorWidget

CifData = DataFactory('cif')  # pylint: disable=invalid-name
StructureData = DataFactory('structure')  # pylint: disable=invalid-name

SYMBOL_RADIUS = {key: covalent_radii[i] for i, key in enumerate(chemical_symbols)}


class StructureManagerWidget(ipw.VBox):
    '''Upload a structure and store it in AiiDA database.

    Attributes:
        structure(Atoms): trait that contains the selected structure. 'None' if no structure is selected.
        structure_node(StructureData, CifData): trait that contains AiiDA structure object
        node_class(str): trait that contains structure_node type (as string).
    '''

    input_structure = Union([Instance(Atoms), Instance(Data)], allow_none=True)
    structure = Union([Instance(Atoms), Instance(Data)], allow_none=True)
    structure_node = Instance(Data, allow_none=True, read_only=True)
    node_class = Unicode()

    SUPPORTED_DATA_FORMATS = {'CifData': 'cif', 'StructureData': 'structure'}

    def __init__(self, importers, viewer=None, editors=None, storable=True, node_class=None, **kwargs):
        """
        Arguments:
            importers(list): list of tuples each containing the displayed name of importer and the
                importer object. Each object should contain 'structure' trait pointing to the imported
                structure. The trait will be linked to 'structure' trait of this class.

            storable(bool): Whether to provide Store button (together with Store format)

            node_class(str): AiiDA node class for storing the structure.
                Possible values: 'StructureData', 'CifData' or None (let the user decide).
                Note: If your workflows require a specific node class, better fix it here.
        """

        # History of modifications
        self.history = []

        # Undo functionality.
        btn_undo = ipw.Button(description='Undo', button_style='success')
        btn_undo.on_click(self.undo)
        self.structure_set_by_undo = False

        # To keep track of last inserted structure object
        self._inserted_structure = None

        # Structure viewer.
        if viewer:
            self.viewer = viewer
        else:
            self.viewer = StructureDataViewer(downloadable=False)
        dlink((self, 'structure'), (self.viewer, 'structure'))

        # Store button.
        self.btn_store = ipw.Button(description='Store in AiiDA', disabled=True)
        self.btn_store.on_click(self.store_structure)

        # Label and description that are stored along with the new structure.
        self.structure_label = ipw.Text(description='Label')
        self.structure_description = ipw.Text(description='Description')

        # Store format selector.
        data_format = ipw.RadioButtons(options=self.SUPPORTED_DATA_FORMATS, description='Data type:')
        link((data_format, 'label'), (self, 'node_class'))

        # Store button, store class selector, description.
        store_and_description = [self.btn_store] if storable else []

        if node_class is None:
            store_and_description.append(data_format)
        elif node_class in self.SUPPORTED_DATA_FORMATS:
            self.node_class = node_class
        else:
            raise ValueError("Unknown data format '{}'. Options: {}".format(node_class,
                                                                            list(self.SUPPORTED_DATA_FORMATS.keys())))
        self.output = ipw.HTML('')

        children = [
            self._structure_importers(importers), self.viewer,
            ipw.HBox(store_and_description + [self.structure_label, self.structure_description])
        ]

        structure_editors = self._structure_editors(editors)
        if structure_editors:
            structure_editors = ipw.VBox([btn_undo, structure_editors])
            accordion = ipw.Accordion([structure_editors])
            accordion.selected_index = None
            accordion.set_title(0, 'Edit Structure')
            children += [accordion]

        super().__init__(children=children + [self.output], **kwargs)

    def _structure_importers(self, importers):
        """Preparing structure importers."""
        if not importers:
            raise ValueError("The parameter importers should contain a list (or tuple) of "
                             "importers, got a falsy object.")

        # If there is only one importer - no need to make tabs.
        if len(importers) == 1:
            # Assigning a function which will be called when importer provides a structure.
            dlink((importers[0], 'structure'), (self, 'input_structure'))
            return importers[0]

        # Otherwise making one tab per importer.
        importers_tab = ipw.Tab()
        importers_tab.children = [i for i in importers]  # One importer per tab.
        for i, importer in enumerate(importers):
            # Labeling tabs.
            importers_tab.set_title(i, importer.title)
            dlink((importer, 'structure'), (self, 'input_structure'))
        return importers_tab

    def _structure_editors(self, editors):
        """Preparing structure editors."""
        if editors and len(editors) == 1:
            link((editors[0], 'structure'), (self, 'structure'))
            if editors[0].has_trait('selection'):
                link((editors[0], 'selection'), (self.viewer, 'selection'))
            if editors[0].has_trait('camera_orientation'):
                dlink((self.viewer._viewer, '_camera_orientation'), (editors[0], 'camera_orientation'))  # pylint: disable=protected-access
            return editors[0]

        # If more than one editor was defined.
        if editors:
            editors_tab = ipw.Tab()
            editors_tab.children = tuple(editors)
            for i, editor in enumerate(editors):
                editors_tab.set_title(i, editor.title)
                link((editor, 'structure'), (self, 'structure'))
                if editor.has_trait('selection'):
                    link((editor, 'selection'), (self.viewer, 'selection'))
                if editor.has_trait('camera_orientation'):
                    dlink((self.viewer._viewer, '_camera_orientation'), (editor, 'camera_orientation'))  # pylint: disable=protected-access
            return editors_tab
        return None

    def store_structure(self, _=None):
        """Stores the structure in AiiDA database."""

        if self.structure_node is None:
            return
        if self.structure_node.is_stored:
            self.output.value = "Already stored in AiiDA [{}], skipping...".format(self.structure_node)
            return
        self.btn_store.disabled = True
        self.structure_node.label = self.structure_label.value
        self.structure_label.disabled = True
        self.structure_node.description = self.structure_description.value
        self.structure_description.disabled = True

        if isinstance(self.input_structure, Data) and self.input_structure.is_stored:

            # Make a link between self.input_structure and self.structure_node
            @calcfunction
            def user_modifications(source_structure):  # pylint: disable=unused-argument
                return self.structure_node

            structure_node = user_modifications(self.input_structure)

        else:
            structure_node = self.structure_node.store()
        self.output.value = "Stored in AiiDA [{}]".format(structure_node)

    def undo(self, _):
        """Undo modifications."""
        self.structure_set_by_undo = True
        if self.history:
            self.history = self.history[:-1]
            if self.history:
                self.structure = self.history[-1]
            else:
                self.input_structure = None
        self.structure_set_by_undo = False

    @staticmethod
    @default('node_class')
    def _default_node_class():
        return 'StructureData'

    @observe('node_class')
    def _change_structure_node(self, _=None):
        with self.hold_trait_notifications():
            self._sync_structure_node()

    def _sync_structure_node(self):
        """Synchronize the structure_node trait using the currently provided info."""
        if len(self.history) > 1:
            # There are some modifications, so converting from ASE.
            structure_node = self._convert_to_structure_node(self.structure)
        else:
            structure_node = self._convert_to_structure_node(self.input_structure)
        self.set_trait('structure_node', structure_node)

    def _convert_to_structure_node(self, structure):
        """Convert structure of any type to the StructureNode object."""
        if structure is None:
            return None
        StructureNode = DataFactory(self.SUPPORTED_DATA_FORMATS[self.node_class])  # pylint: disable=invalid-name

        # If the input_structure trait is set to Atoms object, structure node must be created from it.
        if isinstance(structure, Atoms):
            return StructureNode(ase=structure)

        # If the input_structure trait is set to AiiDA node, check what type
        if isinstance(structure, Data):
            # Transform the structure to the StructureNode if needed.
            if isinstance(structure, StructureNode):
                return structure

        # Using self.structure, as it was already converted to the ASE Atoms object.
        return StructureNode(ase=self.structure)

    @observe('structure_node')
    def _observe_structure_node(self, change):
        """Modify structure label and description when a new structure is provided."""
        struct = change['new']
        if struct is None:
            self.btn_store.disabled = True
            self.structure_label.value = ''
            self.structure_label.disabled = True
            self.structure_description.value = ''
            self.structure_description.disabled = True
            return
        if struct.is_stored:
            self.btn_store.disabled = True
            self.structure_label.value = struct.label
            self.structure_label.disabled = True
            self.structure_description.value = struct.description
            self.structure_description.disabled = True
        else:
            self.btn_store.disabled = False
            self.structure_label.value = self.structure.get_chemical_formula()
            self.structure_label.disabled = False
            self.structure_description.value = ''
            self.structure_description.disabled = False

    @observe('input_structure')
    def _observe_input_structure(self, change):
        """Returns ASE atoms object and sets structure_node trait."""
        # If the `input_structure` trait is set to Atoms object, then the `structure` trait should be set to it as well.
        self.history = []

        if isinstance(change['new'], Atoms):
            self.structure = change['new']

        # If the `input_structure` trait is set to AiiDA node, then the `structure` trait should
        # be converted to an ASE Atoms object.
        elif isinstance(change['new'], CifData):  # Special treatement of the CifData object
            str_io = io.StringIO(change['new'].get_content())
            self.structure = ase.io.read(str_io, format='cif', reader='ase', store_tags=True)
        elif isinstance(change['new'], StructureData):
            self.structure = change['new'].get_ase()

        else:
            self.structure = None

    @observe('structure')
    def _structure_changed(self, change=None):
        """Perform some operations that depend on the value of `structure` trait.

        This function enables/disables `btn_store` widget if structure is provided/set to None.
        Also, the function sets `structure_node` trait to the selected node type.
        """
        if not self.structure_set_by_undo:
            self.history.append(change['new'])

        # If structure trait was set to None, structure_node should become None as well.
        if self.structure is None:
            self.set_trait('structure_node', None)
            self.btn_store.disabled = True
            return

        self.btn_store.disabled = False
        with self.hold_trait_notifications():
            self._sync_structure_node()


class StructureUploadWidget(ipw.VBox):
    """Class that allows to upload structures from user's computer."""
    structure = Union([Instance(Atoms), Instance(Data)], allow_none=True)

    def __init__(self, title='', description="Upload Structure"):
        self.title = title
        self.file_upload = ipw.FileUpload(description=description, multiple=False, layout={'width': 'initial'})
        supported_formats = ipw.HTML(
            """<a href="https://wiki.fysik.dtu.dk/ase/_modules/ase/io/formats.html" target="_blank">
        Supported structure formats
        </a>""")
        self.file_upload.observe(self._on_file_upload, names='value')
        super().__init__(children=[self.file_upload, supported_formats])

    def _validate_and_fix_ase_cell(self, ase_structure, vacuum_ang=10.0):
        """
        Checks if the ase Atoms object has a cell set,
        otherwise sets it to bounding box plus specified "vacuum" space
        """
        cell = ase_structure.cell

        if (np.linalg.norm(cell[0]) < 0.1 or np.linalg.norm(cell[1]) < 0.1 or np.linalg.norm(cell[2]) < 0.1):
            # if any of the cell vectors is too short, consider it faulty
            # set cell as bounding box + vacuum_ang
            bbox = np.ptp(ase_structure.positions, axis=0)
            new_structure = ase_structure.copy()
            new_structure.cell = bbox + vacuum_ang
            return new_structure
        return ase_structure

    def _on_file_upload(self, change=None):
        """When file upload button is pressed."""
        for fname, item in change['new'].items():
            frmt = fname.split('.')[-1]
            if frmt == 'cif':
                self.structure = CifData(file=io.BytesIO(item['content']))
            else:
                self.structure = self._validate_and_fix_ase_cell(
                    get_ase_from_file(io.StringIO(item['content'].decode()), format=frmt))
            self.file_upload.value.clear()
            break


class StructureExamplesWidget(ipw.VBox):
    """Class to provide example structures for selection."""
    structure = Instance(Atoms, allow_none=True)

    def __init__(self, examples, title='', **kwargs):
        self.title = title
        self.on_structure_selection = lambda structure_ase, name: None
        self._select_structure = ipw.Dropdown(options=self.get_example_structures(examples))
        self._select_structure.observe(self._on_select_structure, names=['value'])
        super().__init__(children=[self._select_structure], **kwargs)

    @staticmethod
    def get_example_structures(examples):
        """Get the list of example structures."""
        if not isinstance(examples, list):
            raise ValueError("parameter examples should be of type list, {} given".format(type(examples)))
        return [("Select structure", False)] + examples

    def _on_select_structure(self, change):  # pylint: disable=unused-argument
        """When structure is selected."""

        self.structure = get_ase_from_file(self._select_structure.value) if self._select_structure.value else None

    @default('structure')
    def _default_structure(self):
        return None


class StructureBrowserWidget(ipw.VBox):
    """Class to query for structures stored in the AiiDA database."""
    structure = Union([Instance(Atoms), Instance(Data)], allow_none=True)

    def __init__(self, title=''):
        self.title = title

        # Structure objects we want to query for.
        self.query_structure_type = (DataFactory('structure'), DataFactory('cif'))

        # Extracting available process labels.
        qbuilder = QueryBuilder().append((CalcJobNode, WorkChainNode), project="label")
        self.drop_label = ipw.Dropdown(options=sorted({'All'}.union({i[0] for i in qbuilder.iterall() if i[0]})),
                                       value='All',
                                       description='Process Label',
                                       disabled=True,
                                       style={'description_width': '120px'},
                                       layout={'width': '50%'})
        self.drop_label.observe(self.search, names='value')

        # Disable process labels selection if we are not looking for the calculated structures.
        def disable_drop_label(change):
            self.drop_label.disabled = not change['new'] == 'calculated'

        # Select structures kind.
        self.mode = ipw.RadioButtons(options=['all', 'uploaded', 'edited', 'calculated'], layout={'width': '25%'})
        self.mode.observe(self.search, names='value')
        self.mode.observe(disable_drop_label, names='value')

        # Date range.
        # Note: there is Date picker widget, but it currently does not work in Safari:
        # https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20List.html#Date-picker
        date_text = ipw.HTML(value='<p>Select the date range:</p>')
        self.start_date_widget = ipw.Text(value='', description='From: ', style={'description_width': '120px'})
        self.end_date_widget = ipw.Text(value='', description='To: ')

        # Search button.
        btn_search = ipw.Button(description='Search',
                                button_style='info',
                                layout={
                                    'width': 'initial',
                                    'margin': '2px 0 0 2em'
                                })
        btn_search.on_click(self.search)

        age_selection = ipw.VBox(
            [date_text, ipw.HBox([self.start_date_widget, self.end_date_widget, btn_search])],
            layout={
                'border': '1px solid #fafafa',
                'padding': '1em'
            })

        h_line = ipw.HTML('<hr>')
        box = ipw.VBox([age_selection, h_line, ipw.HBox([self.mode, self.drop_label])])

        self.results = ipw.Dropdown(layout={'width': '900px'})
        self.results.observe(self._on_select_structure)
        self.search()
        super().__init__([box, h_line, self.results])

    def preprocess(self):
        """Search structures in AiiDA database and add formula extra to them."""

        queryb = QueryBuilder()
        queryb.append(self.query_structure_type, filters={'extras': {'!has_key': 'formula'}})
        for item in queryb.all():  # iterall() would interfere with set_extra()
            try:
                formula = item[0].get_formula()
            except AttributeError:
                # Slow part.
                formula = item[0].get_ase().get_chemical_formula()
            item[0].set_extra("formula", formula)

    def search(self, _=None):
        """Launch the search of structures in AiiDA database."""
        self.preprocess()

        qbuild = QueryBuilder()

        # If the date range is valid, use it for the search
        try:
            start_date = datetime.datetime.strptime(self.start_date_widget.value, '%Y-%m-%d')
            end_date = datetime.datetime.strptime(self.end_date_widget.value, '%Y-%m-%d') + datetime.timedelta(hours=24)

        # Otherwise revert to the standard (i.e. last 7 days)
        except ValueError:
            start_date = datetime.datetime.now() - datetime.timedelta(days=7)
            end_date = datetime.datetime.now() + datetime.timedelta(hours=24)

            self.start_date_widget.value = start_date.strftime('%Y-%m-%d')
            self.end_date_widget.value = end_date.strftime('%Y-%m-%d')

        filters = {}
        filters['ctime'] = {'and': [{'>': start_date}, {'<=': end_date}]}

        if self.mode.value == "uploaded":
            qbuild2 = QueryBuilder().append(self.query_structure_type, project=["id"],
                                            tag='structures').append(Node, with_outgoing='structures')
            processed_nodes = [n[0] for n in qbuild2.all()]
            if processed_nodes:
                filters['id'] = {"!in": processed_nodes}
            qbuild.append(self.query_structure_type, filters=filters)

        elif self.mode.value == "calculated":
            if self.drop_label.value == 'All':
                qbuild.append((CalcJobNode, WorkChainNode), tag='calcjobworkchain')
            else:
                qbuild.append((CalcJobNode, WorkChainNode),
                              filters={'label': self.drop_label.value},
                              tag='calcjobworkchain')
            qbuild.append(self.query_structure_type, with_incoming='calcjobworkchain', filters=filters)

        elif self.mode.value == "edited":
            qbuild.append(CalcFunctionNode)
            qbuild.append(self.query_structure_type, with_incoming=CalcFunctionNode, filters=filters)

        elif self.mode.value == "all":
            qbuild.append(self.query_structure_type, filters=filters)

        qbuild.order_by({self.query_structure_type: {'ctime': 'desc'}})
        matches = {n[0] for n in qbuild.iterall()}
        matches = sorted(matches, reverse=True, key=lambda n: n.ctime)

        options = OrderedDict()
        options["Select a Structure ({} found)".format(len(matches))] = False

        for mch in matches:
            label = "PK: {}".format(mch.id)
            label += " | " + mch.ctime.strftime("%Y-%m-%d %H:%M")
            label += " | " + mch.get_extra("formula")
            label += " | " + mch.node_type.split('.')[-2]
            label += " | " + mch.label
            label += " | " + mch.description
            options[label] = mch

        self.results.options = options

    def _on_select_structure(self, _=None):
        self.structure = self.results.value or None


class SmilesWidget(ipw.VBox):
    """Conver SMILES into 3D structure."""
    structure = Instance(Atoms, allow_none=True)

    SPINNER = """<i class="fa fa-spinner fa-pulse" style="color:red;" ></i>"""

    def __init__(self, title=''):
        self.title = title
        try:
            import openbabel  # pylint: disable=unused-import
        except ImportError:
            super().__init__(
                [ipw.HTML("The SmilesWidget requires the OpenBabel library, "
                          "but the library was not found.")])
            return

        self.smiles = ipw.Text()
        self.create_structure_btn = ipw.Button(description="Generate molecule", button_style='info')
        self.create_structure_btn.on_click(self._on_button_pressed)
        self.output = ipw.HTML("")
        super().__init__([self.smiles, self.create_structure_btn, self.output])

    @staticmethod
    def pymol_2_ase(pymol):
        """Convert pymol object into ASE Atoms."""

        species = [chemical_symbols[atm.atomicnum] for atm in pymol.atoms]
        pos = np.asarray([atm.coords for atm in pymol.atoms])
        pca = PCA(n_components=3)
        posnew = pca.fit_transform(pos)
        atoms = Atoms(species, positions=posnew, pbc=True, cell=np.ptp(posnew, axis=0) + 10)
        atoms.center()
        return atoms

    def _optimize_mol(self, mol):
        """Optimize a molecule using force field (needed for complex SMILES)."""

        # Note, the pybel module imported below comes together with openbabel package. Do not confuse it with
        # pybel package available on PyPi: https://pypi.org/project/pybel/
        import pybel  # pylint:disable=import-error

        self.output.value = "Screening possible conformers {}".format(self.SPINNER)  #font-size:20em;

        f_f = pybel._forcefields["uff"]  # pylint: disable=protected-access
        if not f_f.Setup(mol.OBMol):
            f_f = pybel._forcefields["mmff94"]  # pylint: disable=protected-access
            if not f_f.Setup(mol.OBMol):
                self.output.value = "Cannot set up forcefield"
                return

        # Initial cleanup before the weighted search.
        f_f.Setup(mol.OBMol)
        f_f.SteepestDescent(5000, 1.0e-9)
        f_f.GetCoordinates(mol.OBMol)
        self.output.value = ""

    def _on_button_pressed(self, change):  # pylint: disable=unused-argument
        """Convert SMILES to ase structure when button is pressed."""
        self.output.value = ""

        # Note, the pybel module imported below comes together with openbabel package. Do not confuse it with
        # pybel package available on PyPi: https://pypi.org/project/pybel/
        import pybel  # pylint:disable=import-error

        if not self.smiles.value:
            return

        mol = pybel.readstring("smiles", self.smiles.value)
        self.output.value = """SMILES to 3D conversion {}""".format(self.SPINNER)
        mol.make3D()
        mol.addh()

        pybel._builder.Build(mol.OBMol)  # pylint: disable=protected-access

        self._optimize_mol(mol)
        self.structure = self.pymol_2_ase(mol)

    @default('structure')
    def _default_structure(self):
        return None


class BasicStructureEditor(ipw.VBox):  # pylint: disable=too-many-instance-attributes
    """Widget that allows for the basic structure editing."""

    structure = Instance(Atoms, allow_none=True)
    selection = List(Int)
    camera_orientation = List()

    def __init__(self, title=''):

        self.title = title
        # Define action vector.
        self.axis_p1 = ipw.Text(description='Starting point', value='0 0 0', layout={'width': 'initial'})
        self.axis_p2 = ipw.Text(description='Ending point', value='0 0 1', layout={'width': 'initial'})
        btn_def_atom1 = ipw.Button(description='From selection', layout={'width': 'initial'})
        btn_def_atom1.on_click(self.def_axis_p1)
        btn_def_atom2 = ipw.Button(description='From selection', layout={'width': 'initial'})
        btn_def_atom2.on_click(self.def_axis_p2)
        btn_get_from_camera = ipw.Button(description='Perp. to screen',
                                         button_style='warning',
                                         layout={'width': 'initial'})
        btn_get_from_camera.on_click(self.def_perpendicular_to_screen)

        # Define action point.
        self.point = ipw.Text(description='Action point', value='0 0 0', layout={'width': 'initial'})
        btn_def_pnt = ipw.Button(description='From selection', layout={'width': 'initial'})
        btn_def_pnt.on_click(self.def_point)

        # Move atoms.
        btn_move_dr = ipw.Button(description='Move', layout={'width': 'initial'})
        btn_move_dr.on_click(self.translate_dr)
        self.displacement = ipw.FloatText(description=r'Move along action vector by $\Delta=$ ',
                                          value=1,
                                          step=0.1,
                                          style={'description_width': 'initial'},
                                          layout={'width': 'initial'})

        btn_move_dxyz = ipw.Button(description='Move by XYZ', layout={'width': 'initial'})
        btn_move_dxyz.on_click(self.translate_dxdydz)
        btn_move_to_xyz = ipw.Button(description='Move to XYZ', layout={'width': 'initial'})
        btn_move_to_xyz.on_click(self.translate_to_xyz)
        self.dxyz = ipw.Text(description='XYZ move:',
                             value='0 0 0',
                             style={'description_width': 'initial'},
                             layout={
                                 'width': 'initial',
                                 'margin': '0px 0px 0px 20px'
                             })

        # Rotate atoms.
        btn_rotate = ipw.Button(description='Rotate', layout={'width': '10%'})
        btn_rotate.on_click(self.rotate)
        self.phi = ipw.FloatText(description='Rotate around the action vector which starts from the action point',
                                 value=0,
                                 step=5,
                                 style={'description_width': 'initial'},
                                 layout={'width': 'initial'})

        # Atoms selection.
        self.element = ipw.Dropdown(
            description="Select element",
            options=chemical_symbols[1:],
            value="H",
            style={'description_width': 'initial'},
            layout={'width': 'initial'},
        )

        def disable_element(_=None):
            if self.ligand.value == 0:
                self.element.disabled = False
            else:
                self.element.disabled = True

        # Ligand selection.
        self.ligand = LigandSelectorWidget()
        self.ligand.observe(disable_element, names='value')

        # Add atom.
        btn_add = ipw.Button(description='Add to selected', layout={'width': 'initial'})
        btn_add.on_click(self.add)
        self.bond_length = ipw.FloatText(description="Bond lenght.", value=1.0, layout={'width': '140px'})
        use_covalent_radius = ipw.Checkbox(
            value=True,
            description='Use covalent radius',
            style={'description_width': 'initial'},
        )
        link((use_covalent_radius, 'value'), (self.bond_length, 'disabled'))

        # Modify atom.
        btn_modify = ipw.Button(description='Modify selected', button_style='warning', layout={'width': 'initial'})
        btn_modify.on_click(self.mod_element)

        # Remove atom.
        btn_remove = ipw.Button(description='Remove selected', button_style='danger', layout={'width': 'initial'})
        btn_remove.on_click(self.remove)

        # Automatically clear selection after point definition
        self.autoclear_selection = ipw.Checkbox(description='Clear selection after pressing "From seletion"',
                                                value=True,
                                                style={'description_width': 'initial'})

        super().__init__(children=[
            ipw.HTML("<b>Action vector and point:</b>", layout={'margin': '20px 0px 10px 0px'}),
            ipw.HBox([self.axis_p1, btn_def_atom1, self.axis_p2, btn_def_atom2, btn_get_from_camera],
                     layout={'margin': '0px 0px 0px 20px'}),
            ipw.HBox([self.point, btn_def_pnt, self.autoclear_selection], layout={'margin': '0px 0px 0px 20px'}),
            ipw.HTML("<b>Move atom(s):</b>", layout={'margin': '20px 0px 10px 0px'}),
            ipw.HBox([self.displacement, btn_move_dr, self.dxyz, btn_move_dxyz, btn_move_to_xyz],
                     layout={'margin': '0px 0px 0px 20px'}),
            ipw.HBox([self.phi, btn_rotate], layout={'margin': '0px 0px 0px 20px'}),
            ipw.HTML("<b>Modify atom(s):</v>", layout={'margin': '20px 0px 10px 0px'}),
            ipw.HBox([self.element, self.ligand], layout={'margin': '0px 0px 0px 20px'}),
            ipw.HBox([
                btn_modify,
                btn_add,
                self.bond_length,
                use_covalent_radius,
            ],
                     layout={'margin': '0px 0px 0px 20px'}),
            ipw.HBox([btn_remove], layout={'margin': '0px 0px 0px 20px'}),
        ])

    def str2vec(self, string):
        return np.array(list(map(float, string.split())))

    def vec2str(self, vector):
        return str(round(vector[0], 2)) + ' ' + str(round(vector[1], 2)) + ' ' + str(round(vector[2], 2))

    def sel2com(self):
        """Get center of mass of the selection."""
        if self.selection:
            com = np.average(self.structure[self.selection].get_positions(), axis=0)
        else:
            com = [0, 0, 0]

        return com

    @property
    def action_vector(self):
        """Define the action vector."""
        normal = self.str2vec(self.axis_p2.value) - self.str2vec(self.axis_p1.value)
        return normal / np.linalg.norm(normal)

    def def_point(self, _=None):
        """Define the action point."""
        self.point.value = self.vec2str(self.sel2com())
        if self.autoclear_selection.value:
            self.selection = list()

    def def_axis_p1(self, _=None):
        """Define the first point of axis."""
        self.axis_p1.value = self.vec2str(self.sel2com())
        if self.autoclear_selection.value:
            self.selection = list()

    def def_axis_p2(self, _=None):
        """Define the second point of axis."""
        com = np.average(self.structure[self.selection].get_positions(), axis=0) if self.selection else [0, 0, 1]
        self.axis_p2.value = self.vec2str(com)
        if self.autoclear_selection.value:
            self.selection = list()

    def def_perpendicular_to_screen(self, _=None):
        """Define a normalized vector perpendicular to the screen."""
        cmr = self.camera_orientation
        if cmr:
            self.axis_p1.value = "0 0 0"
            versor = np.array([-cmr[2], -cmr[6], -cmr[10]]) / np.linalg.norm([-cmr[2], -cmr[6], -cmr[10]])
            self.axis_p2.value = self.vec2str(versor.tolist())

    def translate_dr(self, _=None):
        """Translate by dr along the selected vector."""
        atoms = self.structure.copy()

        selection = self.selection

        atoms.positions[self.selection] += np.array(self.action_vector * self.displacement.value)

        self.structure = atoms
        self.selection = selection

    def translate_dxdydz(self, _=None):
        """Translate by the selected XYZ delta."""
        selection = self.selection
        atoms = self.structure.copy()

        # The action.
        atoms.positions[self.selection] += np.array(self.str2vec(self.dxyz.value))

        self.structure = atoms
        self.selection = selection

    def translate_to_xyz(self, _=None):
        """Translate to the selected XYZ position."""
        selection = self.selection
        atoms = self.structure.copy()

        # The action.
        geo_center = np.average(self.structure[self.selection].get_positions(), axis=0)
        atoms.positions[self.selection] += self.str2vec(self.dxyz.value) - geo_center

        self.structure = atoms
        self.selection = selection

    def rotate(self, _=None):
        """Rotate atoms around selected point in space and vector."""

        selection = self.selection
        atoms = self.structure.copy()

        # The action.
        rotated_subset = atoms[self.selection]
        vec = self.str2vec(self.vec2str(self.action_vector))
        center = self.str2vec(self.point.value)
        rotated_subset.rotate(self.phi.value, v=vec, center=center, rotate_cell=False)
        atoms.positions[list(self.selection)] = rotated_subset.positions

        self.structure = atoms
        self.selection = selection

    def mod_element(self, _=None):
        """Modify selected atoms into the given element."""
        atoms = self.structure.copy()
        selection = self.selection

        if self.ligand.value == 0:
            for idx in self.selection:
                new = Atom(self.element.value)
                atoms[idx].mass = new.mass
                atoms[idx].magmom = new.magmom
                atoms[idx].momentum = new.momentum
                atoms[idx].symbol = new.symbol
                atoms[idx].tag = new.tag
                atoms[idx].charge = new.charge
        else:
            initial_ligand = self.ligand.rotate(align_to=self.action_vector, remove_anchor=True)
            for idx in self.selection:
                position = self.structure.positions[idx].copy()
                lgnd = initial_ligand.copy()
                lgnd.translate(position)
                atoms += lgnd

        self.structure = atoms
        self.selection = selection

    def add(self, _=None):
        """Add atoms."""
        atoms = self.structure.copy()
        selection = self.selection

        if self.ligand.value == 0:
            initial_ligand = Atoms([Atom(self.element.value, [0, 0, 0])])
            rad = SYMBOL_RADIUS[self.element.value]
        else:
            initial_ligand = self.ligand.rotate(align_to=self.action_vector)
            rad = SYMBOL_RADIUS[self.ligand.anchoring_atom]

        for idx in self.selection:
            # It is important to copy, otherwise the initial structure will be modified
            position = self.structure.positions[idx].copy()
            lgnd = initial_ligand.copy()

            if self.bond_length.disabled:
                lgnd.translate(position + self.action_vector * (SYMBOL_RADIUS[self.structure.symbols[idx]] + rad))
            else:
                lgnd.translate(position + self.action_vector * self.bond_length.value)

            atoms += lgnd

        self.structure = atoms
        self.selection = selection

    def remove(self, _):
        """Remove selected atoms."""
        atoms = self.structure.copy()
        del [atoms[self.selection]]
        self.structure = atoms



"""Some utility functions used acrross the repository."""
import more_itertools as mit
from ase.io import read


def valid_arguments(arguments, valid_args):
    """Check whether provided arguments are valid."""
    result = {}
    for key, value in arguments.items():
        if key in valid_args:
            if isinstance(value, (tuple, list)):
                result[key] = '\n'.join(value)
            else:
                result[key] = value
    return result


def predefine_settings(obj, **kwargs):
    """Specify some pre-defined settings."""
    for key, value in kwargs.items():
        if hasattr(obj, key):
            setattr(obj, key, value)
        else:
            raise AttributeError("'{}' object has no attirubte '{}'".format(obj, key))


def get_ase_from_file(fname, format=None):  # pylint: disable=redefined-builtin
    """Get ASE structure object."""
    if format == 'cif':
        traj = read(fname, format=format, index=":", store_tags=True)
    else:
        traj = read(fname, format=format, index=":")
    if not traj:
        print(("Could not read any information from the file {}".format(fname)))
        return False
    if len(traj) > 1:
        print(("Warning: Uploaded file {} contained more than one structure. Selecting the first one.".format(fname)))
    return traj[0]


def find_ranges(iterable):
    """Yield range of consecutive numbers."""
    for group in mit.consecutive_groups(iterable):
        group = list(group)
        if len(group) == 1:
            yield group[0]
        else:
            yield group[0], group[-1]


def list_to_string_range(lst, shift=1):
    """Converts a list like [0, 2, 3, 4] into a string like '1 3..5'.

    Shift used when e.g. for a user interface numbering starts from 1 not from 0"""
    return " ".join([
        str(t + shift) if isinstance(t, int) else "{}..{}".format(t[0] + shift, t[1] + shift)
        for t in find_ranges(sorted(lst))
    ])


def string_range_to_list(strng, shift=-1):
    """Converts a string like '1 3..5' into a list like [0, 2, 3, 4].

    Shift used when e.g. for a user interface numbering starts from 1 not from 0"""
    singles = [int(s) + shift for s in strng.split() if s.isdigit()]
    ranges = [r for r in strng.split() if '..' in r]
    if len(singles) + len(ranges) != len(strng.split()):
        return list(), False
    for rng in ranges:
        try:
            start, end = rng.split('..')
            singles += [i + shift for i in range(int(start), int(end) + 1)]
        except ValueError:
            return list(), False
    return singles, True



"""All functionality needed to setup a computer."""

import os
from os import path
from copy import copy
from subprocess import check_output, call

import pexpect
import shortuuid
import ipywidgets as ipw
from IPython.display import clear_output
from traitlets import Bool, Dict, Instance, Int, Unicode, Union, link, observe, validate

from aiida.common import NotExistent
from aiida.orm import Computer, QueryBuilder, User

from aiida.transports.plugins.ssh import parse_sshconfig

STYLE = {"description_width": "200px"}


class SshComputerSetup(ipw.VBox):
    """Setup password-free access to a computer."""
    setup_counter = Int(0)  # Traitlet to inform other widgets about changes
    hostname = Unicode()
    port = Union([Unicode(), Int()])
    use_proxy = Bool()
    username = Unicode()
    proxy_hostname = Unicode()
    proxy_username = Unicode()

    def __init__(self, **kwargs):
        computer_image = ipw.HTML('<img width="200px" src="./miscellaneous/images/computer.png">')

        # Username.
        inp_username = ipw.Text(description="SSH username:", layout=ipw.Layout(width="350px"), style=STYLE)
        link((inp_username, 'value'), (self, 'username'))

        # Port.
        inp_port = ipw.IntText(description="SSH port:", value=22, layout=ipw.Layout(width="350px"), style=STYLE)
        link((inp_port, 'value'), (self, 'port'))

        # Hostname.
        inp_computer_hostname = ipw.Text(description="Computer hostname:",
                                         layout=ipw.Layout(width="350px"),
                                         style=STYLE)
        link((inp_computer_hostname, 'value'), (self, 'hostname'))

        # Upload private key directly.
        self._inp_password = ipw.Password(description="SSH password:",
                                          layout=ipw.Layout(width="150px"),
                                          style=STYLE,
                                          disabled=False)
        self._inp_private_key = ipw.FileUpload(accept='',
                                               layout=ipw.Layout(width="350px"),
                                               style=STYLE,
                                               description='Private key',
                                               multiple=False,
                                               disabled=True)
        self._verification_mode = ipw.Dropdown(options=['password', 'private_key'],
                                               layout=ipw.Layout(width="350px"),
                                               style=STYLE,
                                               value='password',
                                               description='verification mode:',
                                               disabled=False)
        self._verification_mode.observe(self.on_use_verification_mode_change, names='value')

        # Proxy ssh settings.
        inp_use_proxy = ipw.Checkbox(value=False, description='Use proxy')
        inp_use_proxy.observe(self.on_use_proxy_change, names='value')
        link((inp_use_proxy, 'value'), (self, 'use_proxy'))

        inp_proxy_hostname = ipw.Text(description="Proxy server address:",
                                      layout=ipw.Layout(width="350px"),
                                      style=STYLE)
        link((inp_proxy_hostname, 'value'), (self, 'proxy_hostname'))

        self._use_diff_proxy_username = ipw.Checkbox(value=False,
                                                     description='Use different username and password',
                                                     layout={'width': 'initial'})
        self._use_diff_proxy_username.observe(self.on_use_diff_proxy_username_change, names='value')
        inp_proxy_username = ipw.Text(value='',
                                      description="Proxy server username:",
                                      layout=ipw.Layout(width="350px"),
                                      style=STYLE)
        link((inp_proxy_username, 'value'), (self, 'proxy_username'))

        self._inp_proxy_password = ipw.Password(value='',
                                                description="Proxy server password:",
                                                layout=ipw.Layout(width="138px"),
                                                style=STYLE)

        # Setup ssh button and output.
        btn_setup_ssh = ipw.Button(description="Setup ssh")
        btn_setup_ssh.on_click(self.on_setup_ssh)
        self._setup_ssh_out = ipw.Output()

        # Defining widgets positions.
        computer_ssh_box = ipw.VBox([
            inp_computer_hostname, inp_port, inp_username, self._verification_mode, self._inp_password,
            self._inp_private_key, inp_use_proxy
        ],
                                    layout=ipw.Layout(width="400px"))
        self._proxy_user_password_box = ipw.VBox([inp_proxy_username, self._inp_proxy_password],
                                                 layout={'visibility': 'hidden'})
        self._proxy_ssh_box = ipw.VBox(
            [inp_proxy_hostname, self._use_diff_proxy_username, self._proxy_user_password_box],
            layout={
                'visibility': 'hidden',
                'width': '400px'
            })

        children = [
            ipw.HBox([computer_image, computer_ssh_box, self._proxy_ssh_box]), btn_setup_ssh, self._setup_ssh_out
        ]
        super(SshComputerSetup, self).__init__(children, **kwargs)

    @staticmethod
    def _ssh_keygen():
        """Generate ssh key pair."""
        fname = path.expanduser("~/.ssh/id_rsa")
        if not path.exists(fname):
            print("Creating ssh key pair")
            # returns non-0 if the key pair already exists
            call(["ssh-keygen", "-f", fname, "-t", "rsa", "-b", "4096", "-m", "PEM", "-N", ""])

    def is_host_known(self, hostname=None):
        """Check if the host is known already."""
        if hostname is None:
            hostname = self.hostname
        fname = path.expanduser("~/.ssh/known_hosts")
        if not path.exists(fname):
            return False
        return call(["ssh-keygen", "-F", hostname]) == 0

    def _make_host_known(self, hostname, proxycmd=None):
        """Add host information into known_hosts file."""
        proxycmd = [] if proxycmd is None else proxycmd
        fname = path.expanduser("~/.ssh/known_hosts")
        print(f"Adding keys from {hostname} to {fname}")
        hashes = check_output(proxycmd + ["ssh-keyscan", "-p", str(self.port), "-H", hostname])
        with open(fname, "a") as fobj:
            fobj.write(hashes.decode("utf-8"))

    def can_login(self, silent=False):
        """Check if it is possible to login into the remote host."""
        if self.username is None:  # if I can't find the username - I must fail
            return False
        userhost = self.username + "@" + self.hostname
        if not silent:
            print(f"Trying ssh {userhost} -p {self.port}... ", end='')
        # With BatchMode on, no password prompt or other interaction is attempted,
        # so a connect that requires a password will fail.
        ret = call(["ssh", userhost, "-p", str(self.port), "-o", "BatchMode=yes", "-o", "ConnectTimeout=5", "true"])
        if not silent:
            print("Ok" if ret == 0 else "Failed")
        return ret == 0

    def is_in_config(self):
        """Check if the config file contains host information."""
        fname = path.expanduser("~/.ssh/config")
        if not path.exists(fname):
            return False
        cfglines = open(fname).read().split("\n")
        return "Host " + self.hostname in cfglines

    def _write_ssh_config(self, proxycmd='', private_key_abs_fname=None):
        """Put host information into the config file."""
        fname = path.expanduser("~/.ssh/config")
        print(f"Adding section to {fname}")
        with open(fname, "a") as file:
            file.write(f"Host {self.hostname} \n")
            file.write(f"User {self.username} \n")
            file.write(f"Port {self.port} \n")
            if private_key_abs_fname:
                file.write(f"IdentityFile {private_key_abs_fname} \n")
            if proxycmd:
                file.write(f"ProxyCommand ssh -q -Y {proxycmd} netcat %h %p\n")
            file.write("ServerAliveInterval 5\n")

    @staticmethod
    def _add_private_key(private_key_fname, private_key_content):
        """
        param private_key_fname: string
        param private_key_content: bytes
        """
        fname = path.expanduser(f"~/.ssh/{private_key_fname}")
        if path.exists(fname):
            # if file already exist and have the same content
            with open(fname, "rb") as file:
                content = file.read()
                if content == private_key_content:
                    return path.basename(fname)

            fname = fname + '_' + shortuuid.uuid()
        with open(fname, "wb") as file:
            file.write(private_key_content)

        os.chmod(fname, 0o600)

        return fname

    @staticmethod
    def _send_pubkey(hostname, username, password, proxycmd=''):
        """Send a publick key to a remote host."""
        from pexpect import TIMEOUT
        timeout = 10
        print("Sending public key to {}... ".format(hostname), end='')
        str_ssh = 'ssh-copy-id {}@{}'.format(username, hostname)
        if proxycmd:
            str_ssh += ' -o "ProxyCommand ssh -q -Y ' + proxycmd + ' netcat %h %p\n"'
        child = pexpect.spawn(str_ssh)
        try:
            index = child.expect(
                [
                    "s password:",  # 0
                    "All keys were skipped because they already exist on the remote system",  # 1
                    "ERROR: No identities found",  # 2
                    "Could not resolve hostname",  # 3
                    pexpect.EOF
                ],
                timeout=timeout)  # final
        except TIMEOUT:
            print("Exceeded {} s timeout".format(timeout))
            return False

        possible_output = {
            1: {
                'message': "Keys are already present on the remote machine",
                'status': True
            },
            2: {
                'message': "Failed\nLooks like the key pair is not present in ~/.ssh folder",
                'status': False
            },
            3: {
                'message': "Failed\nUnknown hostname",
                'status': False
            },
        }
        if index == 0:
            child.sendline(password)
            try:
                child.expect("Now try logging into", timeout=5)
            except TIMEOUT:
                print("Failed\nPlease check your username and/or password")
                return False
            print("Ok")
            child.close()
            return True

        if index in possible_output:
            print(possible_output[index]['message'])
            return possible_output[index]['status']

        print("Failed\nUnknown problem")
        print(child.before, child.after)
        child.close()
        return False

    def _configure_proxy(self, password, proxy_password):
        """Configure proxy server."""

        # If proxy IS required.
        if self.use_proxy:
            # Again, some standard checks whether proxy server parameters are provided.
            if self.proxy_hostname is None:  # hostname
                print("Please specify the proxy server hostname")
                return False, ''

            # If proxy username and password should be different from the main computer - they must be provided.
            if self._use_diff_proxy_username.value:

                # Check username.
                if self.proxy_username is not None:
                    proxy_username = self.proxy_username
                else:
                    print("Please specify the proxy server username")
                    return False, ''

                # Check password.
                if not proxy_password.strip():
                    print("Please specify the proxy server password")
                    return False, ''

            # If username and password are the same as for the main computer.
            else:
                proxy_username = self.username
                proxy_password = password

            # Make proxy server known.
            if not self.is_host_known(self.proxy_hostname):
                self._make_host_known(self.proxy_hostname)

            # Finally trying to connect.
            if self._send_pubkey(self.proxy_hostname, proxy_username, proxy_password):
                return True, proxy_username + '@' + self.proxy_hostname

            print(f"Could not send public key to {self.proxy_hostname} (proxy server).")
            return False, ''

        # If proxy is NOT required.
        return True, ''

    def on_setup_ssh(self, change):
        """Setup ssh, password and private key are supported"""
        with self._setup_ssh_out:
            mode = self._verification_mode.value
            self._on_setup_ssh(mode, change)

    def _on_setup_ssh(self, mode, change):  # pylint: disable=unused-argument,too-many-return-statements,too-many-branches
        """ATTENTION: modifying the order of operations in this function can lead to unexpected problems"""
        clear_output()

        # If hostname is not provided - do not do anything.
        if self.hostname is None:  # check hostname
            print("Please specify the computer hostname")
            return

        # Check if password-free access was enabled earlier.
        if self.can_login():
            print("Password-free access is already enabled")
            # It can still happen that password-free access is enabled
            # but host is not present in the config file - fixing this.
            if not self.is_in_config():
                self._write_ssh_config()  # We do not use proxy here, because if computer
                # can be accessed without any info in the config - proxy is not needed.
                self.setup_counter += 1  # Only if config file has changed - increase setup_counter.
            return

        # If couldn't login in the previous step, chek whether all required information is provided.
        if self.username is None:  # check username
            print("Please enter your ssh username")
            return

        if mode == 'password':
            self._ssh_keygen()
            # Temporary passwords.
            password = self.__password
            proxy_password = self.__proxy_password
            private_key_abs_fname = None
            if not password.strip():  # check password
                print("Please enter your ssh password")
                return
        if mode == 'private_key':
            # unwrap private key file and setting temporary private_key content
            private_key_fname, private_key_content = self.__private_key
            proxy_password = self.__proxy_password
            password = None
            if private_key_fname is None:  # check private key file
                print("Please upload your private key file")
                return

            # write private key in ~/.ssh/ and use the name of upload file,
            # if exist, generate random string and append to filename then override current name.
            private_key_abs_fname = self._add_private_key(private_key_fname, private_key_content)

        # get the right commands to access the proxy server (if provided)
        success, proxycmd = self._configure_proxy(password, proxy_password)
        if not success:
            return

        # make host known by ssh on the proxy server
        if not self.is_host_known():
            self._make_host_known(self.hostname, ['ssh'] + [proxycmd] if proxycmd else [])

        if mode == 'password':
            # sending public key to the main host
            if not self._send_pubkey(self.hostname, self.username, password, proxycmd):
                print("Could not send public key to {self.hostname}")
                return

        # modify the ssh config file if necessary
        if not self.is_in_config():
            self._write_ssh_config(proxycmd, private_key_abs_fname)

        # FOR LATER: add a check if new config is different from the current one. If so
        # infrom the user about it.

        # final check
        if self.can_login():
            self.setup_counter += 1
            print("Automatic ssh setup successful :-)")
            return
        print("Automatic ssh setup failed, sorry :-(")
        return

    def on_use_verification_mode_change(self, change):  # pylint: disable=unused-argument
        """which verification mode is chosen."""
        if self._verification_mode.value == 'password':
            self._inp_password.disabled = False
            self._inp_private_key.disabled = True
        if self._verification_mode.value == 'private_key':
            self._inp_password.disabled = True
            self._inp_private_key.disabled = False

    def on_use_proxy_change(self, change):
        """If proxy check-box is clicked."""
        if change['new']:
            self._proxy_ssh_box.layout.visibility = 'visible'
        else:
            self._proxy_ssh_box.layout.visibility = 'hidden'
            self._use_diff_proxy_username.value = False

    def on_use_diff_proxy_username_change(self, change):  # pylint: disable=unused-argument
        """If using different username for proxy check-box is clicked."""
        if change['new']:
            self._proxy_user_password_box.layout.visibility = 'visible'
        else:
            self._proxy_user_password_box.layout.visibility = 'hidden'


# Keep this function only because it might be used later.
# What it does: looks inside .ssh/config file and loads computer setup from
# there (if present)
#     def _get_from_config(self, b):
#         config = parse_sshconfig(self.hostname)
#         if 'user' in config:
#             self._inp_username.value = config['user']
#         else:
#             self._inp_username.value = ''
#         if 'proxycommand' in config:
#             self._use_proxy.value = True
#             proxy = ''.join([ s for s in config['proxycommand'].split() if '@' in s])
#             username, hostname = proxy.split('@')
#             self._inp_proxy_address.value = hostname
#             if username != self.username:
#                 self._use_diff_proxy_username.value = True
#                 self.proxy_username = username
#         else:
#             self._use_proxy.value = False

    @property
    def __password(self):
        """Returning the password and immediately destroying it"""
        passwd = copy(self._inp_password.value)
        self._inp_password.value = ''
        return passwd

    @property
    def __proxy_password(self):
        """Returning the password and immediately destroying it"""
        passwd = copy(self._inp_proxy_password.value)
        self._inp_proxy_password.value = ''
        return passwd

    @property
    def __private_key(self):
        """unwrap private key file and setting filename and file content"""
        if self._inp_private_key.value:
            (fname, _value), *_ = self._inp_private_key.value.items()
            content = copy(_value['content'])
            self._inp_private_key.value.clear()
            self._inp_private_key._counter = 0  # pylint: disable=protected-access
            return fname, content
        return None, None

    @observe('proxy_hostname')
    def _observe_proxy_hostname(self, _=None):
        """Enable 'use proxy' widget if proxy hostname is provided."""
        if self.proxy_hostname:
            self.use_proxy = True

    @observe('proxy_username')
    def _observe_proxy_username(self, _=None):
        """Enable 'use proxy' and 'use different proxy username' widgets if proxy username is provided."""
        if self.proxy_username:
            self.use_proxy = True
            self._use_diff_proxy_username.value = True

    @validate('port')
    def _validate_port(self, provided):  # pylint: disable=no-self-use
        return int(provided['value'])


class AiidaComputerSetup(ipw.VBox):
    """Inform AiiDA about a computer."""
    label = Unicode()
    hostname = Unicode()
    description = Unicode()
    work_dir = Unicode()
    mpirun_command = Unicode()
    mpiprocs_per_machine = Union([Unicode(), Int()])
    prepend_text = Unicode()
    append_text = Unicode()
    transport = Unicode()
    scheduler = Unicode()

    def __init__(self, **kwargs):
        from aiida.transports import Transport
        from aiida.schedulers import Scheduler

        # List of widgets to be displayed.
        inp_computer_name = ipw.Text(value='',
                                     placeholder='Will only be used within AiiDA',
                                     description="AiiDA computer name:",
                                     layout=ipw.Layout(width="500px"),
                                     style=STYLE)
        link((inp_computer_name, 'value'), (self, 'label'))

        # Hostname.
        inp_computer_hostname = ipw.Text(description="Hostname:", layout=ipw.Layout(width="500px"), style=STYLE)
        link((inp_computer_hostname, 'value'), (self, 'hostname'))

        # Computer description.
        inp_computer_description = ipw.Text(value='',
                                            placeholder='No description (yet)',
                                            description="Computer description:",
                                            layout=ipw.Layout(width="500px"),
                                            style=STYLE)
        link((inp_computer_description, 'value'), (self, 'description'))

        # Directory where to run the simulations.
        inp_computer_workdir = ipw.Text(value='/scratch/{username}/aiida_run',
                                        description="Workdir:",
                                        layout=ipw.Layout(width="500px"),
                                        style=STYLE)
        link((inp_computer_workdir, 'value'), (self, 'work_dir'))

        # Mpirun command.
        inp_mpirun_cmd = ipw.Text(value='mpirun -n {tot_num_mpiprocs}',
                                  description="Mpirun command:",
                                  layout=ipw.Layout(width="500px"),
                                  style=STYLE)
        link((inp_mpirun_cmd, 'value'), (self, 'mpirun_command'))

        # Number of CPUs per node.
        inp_computer_ncpus = ipw.IntText(value=12,
                                         step=1,
                                         description='Number of CPU(s) per node:',
                                         layout=ipw.Layout(width="270px"),
                                         style=STYLE)
        link((inp_computer_ncpus, 'value'), (self, 'mpiprocs_per_machine'))

        inp_transport_type = ipw.Dropdown(value='ssh',
                                          options=Transport.get_valid_transports(),
                                          description="Transport type:",
                                          style=STYLE)
        link((inp_transport_type, 'value'), (self, 'transport'))

        # Scheduler.
        inp_scheduler = ipw.Dropdown(value='slurm',
                                     options=Scheduler.get_valid_schedulers(),
                                     description="Scheduler:",
                                     style=STYLE)
        link((inp_scheduler, 'value'), (self, 'scheduler'))

        # Use login shell.
        self._use_login_shell = ipw.Checkbox(value=True, description="Use login shell")

        # Prepend text.
        inp_prepend_text = ipw.Textarea(placeholder='Text to prepend to each command execution',
                                        description='Prepend text:',
                                        layout=ipw.Layout(width="400px"))
        link((inp_prepend_text, 'value'), (self, 'prepend_text'))

        # Append text.
        inp_append_text = ipw.Textarea(placeholder='Text to append to each command execution',
                                       description='Append text:',
                                       layout=ipw.Layout(width="400px"))
        link((inp_append_text, 'value'), (self, 'append_text'))

        # Buttons and outputs.
        btn_setup_comp = ipw.Button(description="Setup computer")
        btn_setup_comp.on_click(self._on_setup_computer)
        btn_test = ipw.Button(description="Test computer")
        btn_test.on_click(self.test)
        self._setup_comp_out = ipw.Output(layout=ipw.Layout(width="500px"))
        self._test_out = ipw.Output(layout=ipw.Layout(width="500px"))

        # Organize the widgets
        children = [
            ipw.HBox([
                ipw.VBox([
                    inp_computer_name,
                    inp_computer_hostname,
                    inp_computer_description,
                    inp_computer_workdir,
                    inp_mpirun_cmd,
                    inp_computer_ncpus,
                    inp_transport_type,
                    inp_scheduler,
                    self._use_login_shell,
                ]),
                ipw.VBox([inp_prepend_text, inp_append_text])
            ]),
            ipw.HBox([btn_setup_comp, btn_test]),
            ipw.HBox([self._setup_comp_out, self._test_out]),
        ]
        super(AiidaComputerSetup, self).__init__(children, **kwargs)

    def _configure_computer(self):
        """Create AuthInfo."""
        print("Configuring '{}'".format(self.label))
        sshcfg = parse_sshconfig(self.hostname)
        authparams = {
            'compress': True,
            'key_filename': os.path.expanduser(sshcfg.get('identityfile', ['~/.ssh/id_rsa'])[0]),
            'gss_auth': False,
            'gss_deleg_creds': False,
            'gss_host': self.hostname,
            'gss_kex': False,
            'key_policy': 'WarningPolicy',
            'load_system_host_keys': True,
            'port': sshcfg.get('port', 22),
            'timeout': 60,
            'use_login_shell': self._use_login_shell.value,
        }
        if 'user' in sshcfg:
            authparams['username'] = sshcfg['user']
        else:
            print(f"SSH username is not provided, please run `verdi computer configure {self.label}` "
                  "from the command line.")
            return
        if 'proxycommand' in sshcfg:
            authparams['proxy_command'] = sshcfg['proxycommand']
        aiidauser = User.objects.get_default()
        from aiida.orm import AuthInfo
        authinfo = AuthInfo(computer=Computer.objects.get(name=self.label), user=aiidauser)
        authinfo.set_auth_params(authparams)
        authinfo.store()
        print(check_output(['verdi', 'computer', 'show', self.label]).decode('utf-8'))

    def _on_setup_computer(self, _=None):
        """When setup computer button is pressed."""
        with self._setup_comp_out:
            clear_output()
            if self.label is None:  # check hostname
                print("Please specify the computer name (for AiiDA)")
                return
            try:
                computer = Computer.objects.get(name=self.label)
                print(f"A computer called {self.label} already exists.")
                return
            except NotExistent:
                pass

            print(f"Creating new computer with name '{self.label}'")
            computer = Computer(name=self.label, hostname=self.hostname, description=self.description)
            computer.set_transport_type(self.transport)
            computer.set_scheduler_type(self.scheduler)
            computer.set_workdir(self.work_dir)
            computer.set_mpirun_command(self.mpirun_command.split())
            computer.set_default_mpiprocs_per_machine(self.mpiprocs_per_machine)
            if self.prepend_text:
                computer.set_prepend_text(self.prepend_text)
            if self.append_text:
                computer.set_append_text(self.append_text)

            computer.store()
            self._configure_computer()

    def test(self, _=None):
        with self._test_out:
            clear_output()
            print(check_output(['verdi', 'computer', 'test', '--print-traceback', self.label]).decode('utf-8'))

    @validate('mpiprocs_per_machine')
    def _validate_mpiprocs_per_machine(self, provided):  # pylint: disable=no-self-use
        return int(provided['value'])


class ComputerDropdown(ipw.VBox):
    """Widget to select a configured computer.

    Attributes:
        selected_computer(Unicode or Computer): Trait that points to the selected Computer instance.
            It can be set either to an AiiDA Computer instance or to a computer label (will
            automatically be replaced by the corresponding Computer instance). It is linked to the
            'value' trait of `self._dropdown` widget.

        computers(Dict): Trait that contains a dictionary (label => Computer instance) for all
        computers found in the AiiDA database. It is linked to the 'options' trait of
        `self._dropdown` widget.

        allow_select_disabled(Bool):  Trait that defines whether to show disabled computers.
    """

    selected_computer = Union([Unicode(), Instance(Computer)], allow_none=True)
    computers = Dict(allow_none=True)
    allow_select_disabled = Bool(False)

    def __init__(self, description='Select computer:', path_to_root='../', **kwargs):
        """Dropdown for configured AiiDA Computers.

        description (str): Text to display before dropdown.

        path_to_root (str): Path to the app's root folder.
        """

        self.output = ipw.HTML()

        self._dropdown = ipw.Dropdown(options={},
                                      value=None,
                                      description=description,
                                      style={'description_width': 'initial'},
                                      disabled=True)
        link((self, 'computers'), (self._dropdown, 'options'))
        link((self._dropdown, 'value'), (self, 'selected_computer'))

        btn_refresh = ipw.Button(description="Refresh", layout=ipw.Layout(width="70px"))
        btn_refresh.on_click(self.refresh)

        self.observe(self.refresh, names='allow_select_disabled')

        self._setup_another = ipw.HTML(
            value=f"""<a href={path_to_root}aiidalab-widgets-base/setup_computer.ipynb target="_blank">
            Setup new computer</a>""")

        children = [ipw.HBox([self._dropdown, btn_refresh, self._setup_another]), self.output]
        self.refresh()
        super().__init__(children=children, **kwargs)

    def _get_computers(self):
        """Get the list of available computers."""

        # Getting the current user.
        user = User.objects.get_default()

        return {
            c[0].name: c[0]
            for c in QueryBuilder().append(Computer).all()
            if c[0].is_user_configured(user) and (self.allow_select_disabled or c[0].is_user_enabled(user))
        }

    def refresh(self, _=None):
        """Refresh the list of configured computers."""
        self.output.value = ''
        with self.hold_trait_notifications():  # pylint: disable=not-context-manager
            self._dropdown.options = self._get_computers()
        if not self.computers:
            self.output.value = "No computers found."
            self._dropdown.disabled = True
        else:
            self._dropdown.disabled = False

        self._dropdown.value = None

    @validate('selected_computer')
    def _validate_selected_computer(self, change):
        """Select computer either by name or by class instance."""
        computer = change['value']
        self.output.value = ''
        if not computer:
            return None
        if isinstance(computer, str):
            if computer in self.computers:
                return self.computers[computer]
            self.output.value = f"""No computer named '<span style="color:red">{computer}</span>'
            was found in your AiiDA database."""

        if isinstance(computer, Computer):
            if computer.name in self.computers:
                return computer
            self.output.value = f"""The computer instance '<span style="color:red">{computer}</span>'
            supplied was not found in your AiiDA database."""
        return None



"""Widgets that allow to query online databases."""
import requests
import ipywidgets as ipw
from traitlets import Bool, Float, Instance, Int, Unicode, default, observe
from ase import Atoms

from optimade_client.query_filter import OptimadeQueryFilterWidget
from optimade_client.query_provider import OptimadeQueryProviderWidget

from aiida.tools.dbimporters.plugins.cod import CodDbImporter


class CodQueryWidget(ipw.VBox):
    '''Query structures in Crystallography Open Database (COD)
    Useful class members:
    :ivar structure(Atoms): trait that contains the selected structure, None if structure is not selected.'''

    structure = Instance(Atoms, allow_none=True)

    def __init__(self, title='', **kwargs):
        description = ipw.HTML("""<h3>Get crystal structures from
    <a href="http://www.crystallography.net">Crystallography Open Database</a></h3>
    <b>Queries by formula</b>
    <br>
    For the queries by formula please adhere to the Hill notation.
    The chemical symbol of an element should be followed by its number in the structure,
    except when there is only one atom.
    When the structure does NOT contain carbon atoms, all the elements are listed alphabetically.
    Example: <i>O2 Si</i>.
    <br>
    In case the structure DOES contain carbon atoms its number following the 'C' symbol is indicated first.
    If hydrogen is also present in the structure then 'H' symbol and its number is indicated second.
    The remaining elements are listed in the alphabetical order. Example: <i>C H4 N2 O</i>.
    <br>
    <b>Queries by the structure id number</b>
    <br>
    For the queries by structure id, plese provide the database id number. Example: <i>1008786</i>
    """)
        self.title = title
        layout = ipw.Layout(width="400px")
        style = {"description_width": "initial"}
        self.inp_elements = ipw.Text(description="",
                                     value="",
                                     placeholder='e.g.: Ni Ti or id number',
                                     layout=layout,
                                     style=style)
        self.btn_query = ipw.Button(description='Query')
        self.query_message = ipw.HTML("Waiting for input...")
        self.drop_structure = ipw.Dropdown(description="",
                                           options=[("select structure", {
                                               "status": False
                                           })],
                                           style=style,
                                           layout=layout)
        self.link = ipw.HTML("Link to the web-page will appear here")
        self.btn_query.on_click(self._on_click_query)
        self.drop_structure.observe(self._on_select_structure, names=['value'])

        children = [
            description,
            ipw.HBox([self.btn_query, self.inp_elements]), self.query_message,
            ipw.HBox([self.drop_structure, self.link])
        ]
        super(CodQueryWidget, self).__init__(children=children, **kwargs)

    @staticmethod
    def _query(idn=None, formula=None):
        """Make the actual query."""
        importer = CodDbImporter()
        if idn is not None:
            return importer.query(id=idn)
        if formula is not None:
            return importer.query(formula=formula)
        return None

    def _on_click_query(self, change):  # pylint: disable=unused-argument
        """Call query when the corresponding button is pressed."""
        structures = [("select structure", {"status": False})]
        idn = None
        formula = None
        self.query_message.value = "Quering the database ... "
        try:
            idn = int(self.inp_elements.value)
        except ValueError:
            formula = str(self.inp_elements.value)

        for entry in self._query(idn=idn, formula=formula):
            try:
                entry_cif = entry.get_cif_node()
                formula = entry_cif.get_ase().get_chemical_formula()
            except:  # pylint: disable=bare-except
                continue
            entry_add = ("{} (id: {})".format(formula, entry.source['id']), {
                "status": True,
                "cif": entry_cif,
                "url": entry.source['uri'],
                "id": entry.source['id'],
            })
            structures.append(entry_add)

        self.query_message.value += "{} structures found".format(len(structures) - 1)
        self.drop_structure.options = structures

    def _on_select_structure(self, change):
        """When a structure was selected."""
        selected = change['new']
        if selected['status'] is False:
            self.structure = None
            return
        self.structure = selected['cif'].get_ase()
        struct_url = selected['url'].split('.cif')[0] + '.html'
        self.link.value = '<a href="{}" target="_blank">COD entry {}</a>'.format(struct_url, selected['id'])

    @default('structure')
    def _default_structure(self):  # pylint: disable=no-self-use
        return None


class OptimadeQueryWidget(ipw.VBox):
    """AiiDAlab-specific OPTIMADE Query widget

    Useful as a widget to integrate with the
    :class:`aiidalab_widgets_base.structures.StructureManagerWidget`,
    embedded into applications.

    NOTE: `embedded` for `OptimadeQueryFilterWidget` was introduced in `optimade-client`
    version 2020.11.5.

    :param embedded: Whether or not to show extra database and provider information.
        When set to `True`, the extra information will be hidden, this is useful
        in situations where the widget is used in a Tab or similar, e.g., for the
        :class:`aiidalab_widgets_base.structures.StructureManagerWidget`.
    :type embedded: bool
    :param title: Title used for Tab header if employed in
        :class:`aiidalab_widgets_base.structures.StructureManagerWidget`.
    :type title: str
    """

    structure = Instance(Atoms, allow_none=True)

    def __init__(
            self,
            embedded: bool = True,
            title: str = None,
            **kwargs,
    ) -> None:
        providers = OptimadeQueryProviderWidget(embedded=embedded)
        filters = OptimadeQueryFilterWidget(embedded=embedded)

        ipw.dlink((providers, 'database'), (filters, 'database'))

        filters.observe(self._update_structure, names='structure')

        self.title = title if title is not None else 'OPTIMADE'
        layout = kwargs.pop('layout') if 'layout' in kwargs else {'width': 'auto', 'height': 'auto'}

        super().__init__(
            children=(providers, filters),
            layout=layout,
            **kwargs,
        )

    def _update_structure(self, change: dict) -> None:
        """New structure chosen"""
        self.structure = change['new'].as_ase if change['new'] else None


class ComputerDatabaseWidget(ipw.HBox):
    """Extract the setup of a known computer from the AiiDA code registry."""
    # Verdi computer setup.
    label = Unicode()
    hostname = Unicode()
    description = Unicode()
    transport = Unicode()
    scheduler = Unicode()
    shebang = Unicode()
    work_dir = Unicode()
    mpirun_command = Unicode()
    mpiprocs_per_machine = Int()
    prepend_text = Unicode()
    append_text = Unicode()
    num_cores_per_mpiproc = Int()
    queue_name = Unicode()

    # Verdi computer configure.
    port = Int()
    allow_agent = Bool()
    safe_interval = Float()
    use_login_shell = Bool()
    proxy_hostname = Unicode()
    proxy_username = Unicode()
    proxy_command = Unicode()

    def __init__(self, **kwargs):
        self.database = {}
        self.update_btn = ipw.Button(description="Pull database")
        self.update_btn.on_click(self.update)
        self.domain = ipw.Dropdown(
            options=[],
            description='Domain',
            disabled=False,
        )
        self.domain.observe(self.update_computers, names=['value', 'options'])

        self.computer = ipw.Dropdown(
            options=[],
            description="Computer:",
            disable=False,
        )
        self.computer.observe(self.update_settings, names=['value', 'options'])

        super().__init__(children=[self.update_btn, self.domain, self.computer], **kwargs)

    def update(self, _=None):
        self.database = requests.get("https://aiidateam.github.io/aiida-code-registry/database.json").json()
        self.domain.options = self.database.keys()

    def update_computers(self, _=None):
        self.computer.options = [key for key in self.database[self.domain.value].keys() if key != "default"]
        self.computer.value = None  # This is a hack to make sure the selected computer appears as selected.
        self.computer.value = self.database[self.domain.value]["default"]

    def update_settings(self, _=None):
        """Read settings from the YAML files and populate self.database with them."""
        if self.domain.value is None or self.computer.value is None:
            return
        computer_settings = self.database[self.domain.value][self.computer.value]
        computer_setup = computer_settings['computer-setup']
        for setting in computer_setup:
            self.set_trait(setting, computer_setup[setting])
        if 'computer-configure' in computer_settings:
            computer_configure = computer_settings['computer-configure']
            for setting in computer_configure:
                self.set_trait(setting, computer_configure[setting])

    @observe('proxy_command')
    def _observe_proxy_command(self, _=None):
        """Extrac username and hostname for connecting to a proxy server."""
        username, hostname = ''.join([w for w in self.proxy_command.split() if '@' in w]).split('@')
        with self.hold_trait_notifications():
            self.proxy_username = username
            self.proxy_hostname = hostname


class CodeDatabaseWidget(ipw.HBox):
    """Extract the setup of a known computer from the AiiDA code registry."""
    label = Unicode()
    description = Unicode()
    input_plugin = Unicode()
    on_computer = Bool()
    remote_abs_path = Unicode()
    computer = Unicode()
    prepend_text = Unicode()
    append_text = Unicode()

    def __init__(self, **kwargs):
        self.database = {}

        # Select domain.
        self.inp_domain = ipw.Dropdown(
            options=[],
            description='Domain',
            disabled=False,
        )
        self.inp_domain.observe(self.update_computers, names=['value', 'options'])

        # Select computer.
        self.inp_computer = ipw.Dropdown(
            options=[],
            description="Computer:",
            disable=False,
        )
        self.inp_computer.observe(self.update_codes, names=['value', 'options'])

        # Select code.
        self.inp_code = ipw.Dropdown(
            options=[],
            description="Code:",
            disable=False,
        )
        self.inp_code.observe(self.update_settings, names=['value', 'options'])

        self.update_btn = ipw.Button(description="Pull database")
        self.update_btn.on_click(self.update)
        super().__init__(children=[self.update_btn, self.inp_domain, self.inp_computer, self.inp_code], **kwargs)

    def update(self, _=None):
        self.database = requests.get("https://aiidateam.github.io/aiida-code-registry/database.json").json()
        self.inp_domain.options = self.database.keys()

    def update_computers(self, _=None):
        self.inp_computer.options = [key for key in self.database[self.inp_domain.value].keys() if key != "default"]
        self.inp_computer.value = None  # This is a hack to make sure the selected computer appears as selected.
        self.inp_computer.value = self.database[self.inp_domain.value]["default"]

    def update_codes(self, _=None):
        """Read settings from the YAML files and populate self.database with them."""
        if self.inp_domain.value is None or self.inp_computer.value is None:
            return
        self.inp_code.options = [
            key for key in self.database[self.inp_domain.value][self.inp_computer.value].keys()
            if key not in ('computer-setup', 'computer-configure')
        ]

    def update_settings(self, _=None):
        """Update code settings."""
        if self.inp_domain.value is None or self.inp_computer.value is None or self.inp_code.value is None:
            return
        settings = self.database[self.inp_domain.value][self.inp_computer.value][self.inp_code.value]
        for key, value in settings.items():
            self.set_trait(key, value)
